"""
Azure (ARM) Network State Module

.. versionadded:: 2019.2.0

.. warning::

    This cloud provider will be removed from Salt in version 3007 in favor of
    the `saltext.azurerm Salt Extension
    <https://github.com/salt-extensions/saltext-azurerm>`_

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example Pillar for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurearm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

    Example states using Azure Resource Manager authentication:

    .. code-block:: jinja

        {% set profile = salt['pillar.get']('azurearm:mysubscription') %}
        Ensure virtual network exists:
            azurearm_network.virtual_network_present:
                - name: my_vnet
                - resource_group: my_rg
                - address_prefixes:
                    - '10.0.0.0/8'
                    - '192.168.0.0/16'
                - dns_servers:
                    - '8.8.8.8'
                - tags:
                    how_awesome: very
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

        Ensure virtual network is absent:
            azurearm_network.virtual_network_absent:
                - name: other_vnet
                - resource_group: my_rg
                - connection_auth: {{ profile }}

"""

import logging
from functools import wraps

import salt.utils.azurearm

__virtualname__ = "azurearm_network"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only make this state available if the azurearm_network module is available.
    """
    if "azurearm_network.check_ip_address_availability" in __salt__:
        return __virtualname__
    return (False, "azurearm_network module could not be loaded")


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Chlorine",
            "The 'azurearm' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3007 in favor of the "
            "saltext.azurerm Salt Extension. "
            "(https://github.com/salt-extensions/saltext-azurerm)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def virtual_network_present(
    name,
    address_prefixes,
    resource_group,
    dns_servers=None,
    tags=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a virtual network exists.

    :param name:
        Name of the virtual network.

    :param resource_group:
        The resource group assigned to the virtual network.

    :param address_prefixes:
        A list of CIDR blocks which can be used by subnets within the virtual network.

    :param dns_servers:
        A list of DNS server addresses.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the virtual network object.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure virtual network exists:
            azurearm_network.virtual_network_present:
                - name: vnet1
                - resource_group: group1
                - address_prefixes:
                    - '10.0.0.0/8'
                    - '192.168.0.0/16'
                - dns_servers:
                    - '8.8.8.8'
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_resource: Ensure resource group exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    vnet = __salt__["azurearm_network.virtual_network_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in vnet:
        tag_changes = __utils__["dictdiffer.deep_diff"](
            vnet.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        dns_changes = set(dns_servers or []).symmetric_difference(
            set(vnet.get("dhcp_options", {}).get("dns_servers", []))
        )
        if dns_changes:
            ret["changes"]["dns_servers"] = {
                "old": vnet.get("dhcp_options", {}).get("dns_servers", []),
                "new": dns_servers,
            }

        addr_changes = set(address_prefixes or []).symmetric_difference(
            set(vnet.get("address_space", {}).get("address_prefixes", []))
        )
        if addr_changes:
            ret["changes"]["address_space"] = {
                "address_prefixes": {
                    "old": vnet.get("address_space", {}).get("address_prefixes", []),
                    "new": address_prefixes,
                }
            }

        if kwargs.get("enable_ddos_protection", False) != vnet.get(
            "enable_ddos_protection"
        ):
            ret["changes"]["enable_ddos_protection"] = {
                "old": vnet.get("enable_ddos_protection"),
                "new": kwargs.get("enable_ddos_protection"),
            }

        if kwargs.get("enable_vm_protection", False) != vnet.get(
            "enable_vm_protection"
        ):
            ret["changes"]["enable_vm_protection"] = {
                "old": vnet.get("enable_vm_protection"),
                "new": kwargs.get("enable_vm_protection"),
            }

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Virtual network {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Virtual network {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "resource_group": resource_group,
                "address_space": {"address_prefixes": address_prefixes},
                "dhcp_options": {"dns_servers": dns_servers},
                "enable_ddos_protection": kwargs.get("enable_ddos_protection", False),
                "enable_vm_protection": kwargs.get("enable_vm_protection", False),
                "tags": tags,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Virtual network {name} would be created."
        ret["result"] = None
        return ret

    vnet_kwargs = kwargs.copy()
    vnet_kwargs.update(connection_auth)

    vnet = __salt__["azurearm_network.virtual_network_create_or_update"](
        name=name,
        resource_group=resource_group,
        address_prefixes=address_prefixes,
        dns_servers=dns_servers,
        tags=tags,
        **vnet_kwargs,
    )

    if "error" not in vnet:
        ret["result"] = True
        ret["comment"] = f"Virtual network {name} has been created."
        return ret

    ret["comment"] = "Failed to create virtual network {}! ({})".format(
        name, vnet.get("error")
    )
    return ret


@_deprecation_message
def virtual_network_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a virtual network does not exist in the resource group.

    :param name:
        Name of the virtual network.

    :param resource_group:
        The resource group assigned to the virtual network.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    vnet = __salt__["azurearm_network.virtual_network_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in vnet:
        ret["result"] = True
        ret["comment"] = f"Virtual network {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Virtual network {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": vnet,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.virtual_network_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Virtual network {name} has been deleted."
        ret["changes"] = {"old": vnet, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete virtual network {name}!"
    return ret


@_deprecation_message
def subnet_present(
    name,
    address_prefix,
    virtual_network,
    resource_group,
    security_group=None,
    route_table=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a subnet exists.

    :param name:
        Name of the subnet.

    :param address_prefix:
        A CIDR block used by the subnet within the virtual network.

    :param virtual_network:
        Name of the existing virtual network to contain the subnet.

    :param resource_group:
        The resource group assigned to the virtual network.

    :param security_group:
        The name of the existing network security group to assign to the subnet.

    :param route_table:
        The name of the existing route table to assign to the subnet.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure subnet exists:
            azurearm_network.subnet_present:
                - name: vnet1_sn1
                - virtual_network: vnet1
                - resource_group: group1
                - address_prefix: '192.168.1.0/24'
                - security_group: nsg1
                - route_table: rt1
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_network: Ensure virtual network exists
                  - azurearm_network: Ensure network security group exists
                  - azurearm_network: Ensure route table exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    snet = __salt__["azurearm_network.subnet_get"](
        name,
        virtual_network,
        resource_group,
        azurearm_log_level="info",
        **connection_auth,
    )

    if "error" not in snet:
        if address_prefix != snet.get("address_prefix"):
            ret["changes"]["address_prefix"] = {
                "old": snet.get("address_prefix"),
                "new": address_prefix,
            }

        nsg_name = None
        if snet.get("network_security_group"):
            nsg_name = snet["network_security_group"]["id"].split("/")[-1]

        if security_group and (security_group != nsg_name):
            ret["changes"]["network_security_group"] = {
                "old": nsg_name,
                "new": security_group,
            }

        rttbl_name = None
        if snet.get("route_table"):
            rttbl_name = snet["route_table"]["id"].split("/")[-1]

        if route_table and (route_table != rttbl_name):
            ret["changes"]["route_table"] = {"old": rttbl_name, "new": route_table}

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Subnet {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Subnet {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "address_prefix": address_prefix,
                "network_security_group": security_group,
                "route_table": route_table,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Subnet {name} would be created."
        ret["result"] = None
        return ret

    snet_kwargs = kwargs.copy()
    snet_kwargs.update(connection_auth)

    snet = __salt__["azurearm_network.subnet_create_or_update"](
        name=name,
        virtual_network=virtual_network,
        resource_group=resource_group,
        address_prefix=address_prefix,
        network_security_group=security_group,
        route_table=route_table,
        **snet_kwargs,
    )

    if "error" not in snet:
        ret["result"] = True
        ret["comment"] = f"Subnet {name} has been created."
        return ret

    ret["comment"] = "Failed to create subnet {}! ({})".format(name, snet.get("error"))
    return ret


@_deprecation_message
def subnet_absent(name, virtual_network, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a virtual network does not exist in the virtual network.

    :param name:
        Name of the subnet.

    :param virtual_network:
        Name of the existing virtual network containing the subnet.

    :param resource_group:
        The resource group assigned to the virtual network.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    snet = __salt__["azurearm_network.subnet_get"](
        name,
        virtual_network,
        resource_group,
        azurearm_log_level="info",
        **connection_auth,
    )

    if "error" in snet:
        ret["result"] = True
        ret["comment"] = f"Subnet {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Subnet {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": snet,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.subnet_delete"](
        name, virtual_network, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Subnet {name} has been deleted."
        ret["changes"] = {"old": snet, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete subnet {name}!"
    return ret


@_deprecation_message
def network_security_group_present(
    name, resource_group, tags=None, security_rules=None, connection_auth=None, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Ensure a network security group exists.

    :param name:
        Name of the network security group.

    :param resource_group:
        The resource group assigned to the network security group.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the network security group object.

    :param security_rules: An optional list of dictionaries representing valid SecurityRule objects. See the
        documentation for the security_rule_present state or security_rule_create_or_update execution module
        for more information on required and optional parameters for security rules. The rules are only
        managed if this parameter is present. When this parameter is absent, implemented rules will not be removed,
        and will merely become unmanaged.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure network security group exists:
            azurearm_network.network_security_group_present:
                - name: nsg1
                - resource_group: group1
                - security_rules:
                  - name: nsg1_rule1
                    priority: 100
                    protocol: tcp
                    access: allow
                    direction: outbound
                    source_address_prefix: virtualnetwork
                    destination_address_prefix: internet
                    source_port_range: '*'
                    destination_port_range: '*'
                  - name: nsg1_rule2
                    priority: 101
                    protocol: tcp
                    access: allow
                    direction: inbound
                    source_address_prefix: internet
                    destination_address_prefix: virtualnetwork
                    source_port_range: '*'
                    destination_port_ranges:
                      - '80'
                      - '443'
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_resource: Ensure resource group exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    nsg = __salt__["azurearm_network.network_security_group_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in nsg:
        tag_changes = __utils__["dictdiffer.deep_diff"](nsg.get("tags", {}), tags or {})
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        if security_rules:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                nsg.get("security_rules", []), security_rules
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"security_rules" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["security_rules"] = comp_ret["changes"]

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Network security group {} is already present.".format(
                name
            )
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Network security group {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "resource_group": resource_group,
                "tags": tags,
                "security_rules": security_rules,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Network security group {name} would be created."
        ret["result"] = None
        return ret

    nsg_kwargs = kwargs.copy()
    nsg_kwargs.update(connection_auth)

    nsg = __salt__["azurearm_network.network_security_group_create_or_update"](
        name=name,
        resource_group=resource_group,
        tags=tags,
        security_rules=security_rules,
        **nsg_kwargs,
    )

    if "error" not in nsg:
        ret["result"] = True
        ret["comment"] = f"Network security group {name} has been created."
        return ret

    ret["comment"] = "Failed to create network security group {}! ({})".format(
        name, nsg.get("error")
    )
    return ret


@_deprecation_message
def network_security_group_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a network security group does not exist in the resource group.

    :param name:
        Name of the network security group.

    :param resource_group:
        The resource group assigned to the network security group.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    nsg = __salt__["azurearm_network.network_security_group_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in nsg:
        ret["result"] = True
        ret["comment"] = f"Network security group {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Network security group {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": nsg,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.network_security_group_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Network security group {name} has been deleted."
        ret["changes"] = {"old": nsg, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete network security group {name}!"
    return ret


@_deprecation_message
def security_rule_present(
    name,
    access,
    direction,
    priority,
    protocol,
    security_group,
    resource_group,
    destination_address_prefix=None,
    destination_port_range=None,
    source_address_prefix=None,
    source_port_range=None,
    description=None,
    destination_address_prefixes=None,
    destination_port_ranges=None,
    source_address_prefixes=None,
    source_port_ranges=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a security rule exists.

    :param name:
        Name of the security rule.

    :param access:
        'allow' or 'deny'

    :param direction:
        'inbound' or 'outbound'

    :param priority:
        Integer between 100 and 4096 used for ordering rule application.

    :param protocol:
        'tcp', 'udp', or '*'

    :param security_group:
        The name of the existing network security group to contain the security rule.

    :param resource_group:
        The resource group assigned to the network security group.

    :param description:
        Optional description of the security rule.

    :param destination_address_prefix:
        The CIDR or destination IP range. Asterix '*' can also be used to match all destination IPs.
        Default tags such as 'VirtualNetwork', 'AzureLoadBalancer' and 'Internet' can also be used.
        If this is an ingress rule, specifies where network traffic originates from.

    :param destination_port_range:
        The destination port or range. Integer or range between 0 and 65535. Asterix '*'
        can also be used to match all ports.

    :param source_address_prefix:
        The CIDR or source IP range. Asterix '*' can also be used to match all source IPs.
        Default tags such as 'VirtualNetwork', 'AzureLoadBalancer' and 'Internet' can also be used.
        If this is an ingress rule, specifies where network traffic originates from.

    :param source_port_range:
        The source port or range. Integer or range between 0 and 65535. Asterix '*'
        can also be used to match all ports.

    :param destination_address_prefixes:
        A list of destination_address_prefix values. This parameter overrides destination_address_prefix
        and will cause any value entered there to be ignored.

    :param destination_port_ranges:
        A list of destination_port_range values. This parameter overrides destination_port_range
        and will cause any value entered there to be ignored.

    :param source_address_prefixes:
        A list of source_address_prefix values. This parameter overrides source_address_prefix
        and will cause any value entered there to be ignored.

    :param source_port_ranges:
        A list of source_port_range values. This parameter overrides source_port_range
        and will cause any value entered there to be ignored.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure security rule exists:
            azurearm_network.security_rule_present:
                - name: nsg1_rule2
                - security_group: nsg1
                - resource_group: group1
                - priority: 101
                - protocol: tcp
                - access: allow
                - direction: inbound
                - source_address_prefix: internet
                - destination_address_prefix: virtualnetwork
                - source_port_range: '*'
                - destination_port_ranges:
                  - '80'
                  - '443'
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_network: Ensure network security group exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    exclusive_params = [
        ("source_port_ranges", "source_port_range"),
        ("source_address_prefixes", "source_address_prefix"),
        ("destination_port_ranges", "destination_port_range"),
        ("destination_address_prefixes", "destination_address_prefix"),
    ]

    for params in exclusive_params:
        # pylint: disable=eval-used
        if not eval(params[0]) and not eval(params[1]):
            ret["comment"] = "Either the {} or {} parameter must be provided!".format(
                params[0], params[1]
            )
            return ret
        # pylint: disable=eval-used
        if eval(params[0]):
            # pylint: disable=eval-used
            if not isinstance(eval(params[0]), list):
                ret["comment"] = f"The {params[0]} parameter must be a list!"
                return ret
            # pylint: disable=exec-used
            exec(f"{params[1]} = None")

    rule = __salt__["azurearm_network.security_rule_get"](
        name,
        security_group,
        resource_group,
        azurearm_log_level="info",
        **connection_auth,
    )

    if "error" not in rule:
        # access changes
        if access.capitalize() != rule.get("access"):
            ret["changes"]["access"] = {"old": rule.get("access"), "new": access}

        # description changes
        if description != rule.get("description"):
            ret["changes"]["description"] = {
                "old": rule.get("description"),
                "new": description,
            }

        # direction changes
        if direction.capitalize() != rule.get("direction"):
            ret["changes"]["direction"] = {
                "old": rule.get("direction"),
                "new": direction,
            }

        # priority changes
        if int(priority) != rule.get("priority"):
            ret["changes"]["priority"] = {"old": rule.get("priority"), "new": priority}

        # protocol changes
        if protocol.lower() != rule.get("protocol", "").lower():
            ret["changes"]["protocol"] = {"old": rule.get("protocol"), "new": protocol}

        # destination_port_range changes
        if destination_port_range != rule.get("destination_port_range"):
            ret["changes"]["destination_port_range"] = {
                "old": rule.get("destination_port_range"),
                "new": destination_port_range,
            }

        # source_port_range changes
        if source_port_range != rule.get("source_port_range"):
            ret["changes"]["source_port_range"] = {
                "old": rule.get("source_port_range"),
                "new": source_port_range,
            }

        # destination_port_ranges changes
        if sorted(destination_port_ranges or []) != sorted(
            rule.get("destination_port_ranges", [])
        ):
            ret["changes"]["destination_port_ranges"] = {
                "old": rule.get("destination_port_ranges"),
                "new": destination_port_ranges,
            }

        # source_port_ranges changes
        if sorted(source_port_ranges or []) != sorted(
            rule.get("source_port_ranges", [])
        ):
            ret["changes"]["source_port_ranges"] = {
                "old": rule.get("source_port_ranges"),
                "new": source_port_ranges,
            }

        # destination_address_prefix changes
        if (destination_address_prefix or "").lower() != rule.get(
            "destination_address_prefix", ""
        ).lower():
            ret["changes"]["destination_address_prefix"] = {
                "old": rule.get("destination_address_prefix"),
                "new": destination_address_prefix,
            }

        # source_address_prefix changes
        if (source_address_prefix or "").lower() != rule.get(
            "source_address_prefix", ""
        ).lower():
            ret["changes"]["source_address_prefix"] = {
                "old": rule.get("source_address_prefix"),
                "new": source_address_prefix,
            }

        # destination_address_prefixes changes
        if sorted(destination_address_prefixes or []) != sorted(
            rule.get("destination_address_prefixes", [])
        ):
            if len(destination_address_prefixes or []) != len(
                rule.get("destination_address_prefixes", [])
            ):
                ret["changes"]["destination_address_prefixes"] = {
                    "old": rule.get("destination_address_prefixes"),
                    "new": destination_address_prefixes,
                }
            else:
                local_dst_addrs, remote_dst_addrs = (
                    sorted(destination_address_prefixes),
                    sorted(rule.get("destination_address_prefixes")),
                )
                for idx, val in enumerate(local_dst_addrs):
                    if val.lower() != remote_dst_addrs[idx].lower():
                        ret["changes"]["destination_address_prefixes"] = {
                            "old": rule.get("destination_address_prefixes"),
                            "new": destination_address_prefixes,
                        }
                        break

        # source_address_prefixes changes
        if sorted(source_address_prefixes or []) != sorted(
            rule.get("source_address_prefixes", [])
        ):
            if len(source_address_prefixes or []) != len(
                rule.get("source_address_prefixes", [])
            ):
                ret["changes"]["source_address_prefixes"] = {
                    "old": rule.get("source_address_prefixes"),
                    "new": source_address_prefixes,
                }
            else:
                local_src_addrs, remote_src_addrs = (
                    sorted(source_address_prefixes),
                    sorted(rule.get("source_address_prefixes")),
                )
                for idx, val in enumerate(local_src_addrs):
                    if val.lower() != remote_src_addrs[idx].lower():
                        ret["changes"]["source_address_prefixes"] = {
                            "old": rule.get("source_address_prefixes"),
                            "new": source_address_prefixes,
                        }
                        break

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Security rule {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Security rule {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "access": access,
                "description": description,
                "direction": direction,
                "priority": priority,
                "protocol": protocol,
                "destination_address_prefix": destination_address_prefix,
                "destination_address_prefixes": destination_address_prefixes,
                "destination_port_range": destination_port_range,
                "destination_port_ranges": destination_port_ranges,
                "source_address_prefix": source_address_prefix,
                "source_address_prefixes": source_address_prefixes,
                "source_port_range": source_port_range,
                "source_port_ranges": source_port_ranges,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Security rule {name} would be created."
        ret["result"] = None
        return ret

    rule_kwargs = kwargs.copy()
    rule_kwargs.update(connection_auth)

    rule = __salt__["azurearm_network.security_rule_create_or_update"](
        name=name,
        access=access,
        description=description,
        direction=direction,
        priority=priority,
        protocol=protocol,
        security_group=security_group,
        resource_group=resource_group,
        destination_address_prefix=destination_address_prefix,
        destination_address_prefixes=destination_address_prefixes,
        destination_port_range=destination_port_range,
        destination_port_ranges=destination_port_ranges,
        source_address_prefix=source_address_prefix,
        source_address_prefixes=source_address_prefixes,
        source_port_range=source_port_range,
        source_port_ranges=source_port_ranges,
        **rule_kwargs,
    )

    if "error" not in rule:
        ret["result"] = True
        ret["comment"] = f"Security rule {name} has been created."
        return ret

    ret["comment"] = "Failed to create security rule {}! ({})".format(
        name, rule.get("error")
    )
    return ret


@_deprecation_message
def security_rule_absent(name, security_group, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a security rule does not exist in the network security group.

    :param name:
        Name of the security rule.

    :param security_group:
        The network security group containing the security rule.

    :param resource_group:
        The resource group assigned to the network security group.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    rule = __salt__["azurearm_network.security_rule_get"](
        name,
        security_group,
        resource_group,
        azurearm_log_level="info",
        **connection_auth,
    )

    if "error" in rule:
        ret["result"] = True
        ret["comment"] = f"Security rule {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Security rule {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": rule,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.security_rule_delete"](
        name, security_group, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Security rule {name} has been deleted."
        ret["changes"] = {"old": rule, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete security rule {name}!"
    return ret


@_deprecation_message
def load_balancer_present(
    name,
    resource_group,
    sku=None,
    frontend_ip_configurations=None,
    backend_address_pools=None,
    load_balancing_rules=None,
    probes=None,
    inbound_nat_rules=None,
    inbound_nat_pools=None,
    outbound_nat_rules=None,
    tags=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a load balancer exists.

    :param name:
        Name of the load balancer.

    :param resource_group:
        The resource group assigned to the load balancer.

    :param sku:
        The load balancer SKU, which can be 'Basic' or 'Standard'.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the load balancer object.

    :param frontend_ip_configurations:
        An optional list of dictionaries representing valid FrontendIPConfiguration objects. A frontend IP
        configuration can be either private (using private IP address and subnet parameters) or public (using a
        reference to a public IP address object). Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``private_ip_address``: The private IP address of the IP configuration. Required if
          'private_ip_allocation_method' is 'Static'.
        - ``private_ip_allocation_method``: The Private IP allocation method. Possible values are: 'Static' and
          'Dynamic'.
        - ``subnet``: Name of an existing subnet inside of which the frontend IP will reside.
        - ``public_ip_address``: Name of an existing public IP address which will be assigned to the frontend IP object.

    :param backend_address_pools:
        An optional list of dictionaries representing valid BackendAddressPool objects. Only the 'name' parameter is
        valid for a BackendAddressPool dictionary. All other parameters are read-only references from other objects
        linking to the backend address pool. Inbound traffic is randomly load balanced across IPs in the backend IPs.

    :param probes:
        An optional list of dictionaries representing valid Probe objects. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``protocol``: The protocol of the endpoint. Possible values are 'Http' or 'Tcp'. If 'Tcp' is specified, a
          received ACK is required for the probe to be successful. If 'Http' is specified, a 200 OK response from the
          specified URI is required for the probe to be successful.
        - ``port``: The port for communicating the probe. Possible values range from 1 to 65535, inclusive.
        - ``interval_in_seconds``: The interval, in seconds, for how frequently to probe the endpoint for health status.
          Typically, the interval is slightly less than half the allocated timeout period (in seconds) which allows two
          full probes before taking the instance out of rotation. The default value is 15, the minimum value is 5.
        - ``number_of_probes``: The number of probes where if no response, will result in stopping further traffic from
          being delivered to the endpoint. This values allows endpoints to be taken out of rotation faster or slower
          than the typical times used in Azure.
        - ``request_path``: The URI used for requesting health status from the VM. Path is required if a protocol is
          set to 'Http'. Otherwise, it is not allowed. There is no default value.

    :param load_balancing_rules:
        An optional list of dictionaries representing valid LoadBalancingRule objects. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``load_distribution``: The load distribution policy for this rule. Possible values are 'Default', 'SourceIP',
          and 'SourceIPProtocol'.
        - ``frontend_port``: The port for the external endpoint. Port numbers for each rule must be unique within the
          Load Balancer. Acceptable values are between 0 and 65534. Note that value 0 enables 'Any Port'.
        - ``backend_port``: The port used for internal connections on the endpoint. Acceptable values are between 0 and
          65535. Note that value 0 enables 'Any Port'.
        - ``idle_timeout_in_minutes``: The timeout for the TCP idle connection. The value can be set between 4 and 30
          minutes. The default value is 4 minutes. This element is only used when the protocol is set to TCP.
        - ``enable_floating_ip``: Configures a virtual machine's endpoint for the floating IP capability required
          to configure a SQL AlwaysOn Availability Group. This setting is required when using the SQL AlwaysOn
          Availability Groups in SQL server. This setting can't be changed after you create the endpoint.
        - ``disable_outbound_snat``: Configures SNAT for the VMs in the backend pool to use the public IP address
          specified in the frontend of the load balancing rule.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the load balancing rule
          object.
        - ``backend_address_pool``: Name of the backend address pool object used by the load balancing rule object.
          Inbound traffic is randomly load balanced across IPs in the backend IPs.
        - ``probe``: Name of the probe object used by the load balancing rule object.

    :param inbound_nat_rules:
        An optional list of dictionaries representing valid InboundNatRule objects. Defining inbound NAT rules on your
        load balancer is mutually exclusive with defining an inbound NAT pool. Inbound NAT pools are referenced from
        virtual machine scale sets. NICs that are associated with individual virtual machines cannot reference an
        Inbound NAT pool. They have to reference individual inbound NAT rules. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the inbound NAT rule
          object.
        - ``protocol``: Possible values include 'Udp', 'Tcp', or 'All'.
        - ``frontend_port``: The port for the external endpoint. Port numbers for each rule must be unique within the
          Load Balancer. Acceptable values range from 1 to 65534.
        - ``backend_port``: The port used for the internal endpoint. Acceptable values range from 1 to 65535.
        - ``idle_timeout_in_minutes``: The timeout for the TCP idle connection. The value can be set between 4 and 30
          minutes. The default value is 4 minutes. This element is only used when the protocol is set to TCP.
        - ``enable_floating_ip``: Configures a virtual machine's endpoint for the floating IP capability required
          to configure a SQL AlwaysOn Availability Group. This setting is required when using the SQL AlwaysOn
          Availability Groups in SQL server. This setting can't be changed after you create the endpoint.

    :param inbound_nat_pools:
        An optional list of dictionaries representing valid InboundNatPool objects. They define an external port range
        for inbound NAT to a single backend port on NICs associated with a load balancer. Inbound NAT rules are created
        automatically for each NIC associated with the Load Balancer using an external port from this range. Defining an
        Inbound NAT pool on your Load Balancer is mutually exclusive with defining inbound NAT rules. Inbound NAT pools
        are referenced from virtual machine scale sets. NICs that are associated with individual virtual machines cannot
        reference an inbound NAT pool. They have to reference individual inbound NAT rules. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the inbound NAT pool
          object.
        - ``protocol``: Possible values include 'Udp', 'Tcp', or 'All'.
        - ``frontend_port_range_start``: The first port number in the range of external ports that will be used to
          provide Inbound NAT to NICs associated with a load balancer. Acceptable values range between 1 and 65534.
        - ``frontend_port_range_end``: The last port number in the range of external ports that will be used to
          provide Inbound NAT to NICs associated with a load balancer. Acceptable values range between 1 and 65535.
        - ``backend_port``: The port used for internal connections to the endpoint. Acceptable values are between 1 and
          65535.

    :param outbound_nat_rules:
        An optional list of dictionaries representing valid OutboundNatRule objects. Valid parameters are:

        - ``name``: The name of the resource that is unique within a resource group.
        - ``frontend_ip_configuration``: Name of the frontend IP configuration object used by the outbound NAT rule
          object.
        - ``backend_address_pool``: Name of the backend address pool object used by the outbound NAT rule object.
          Outbound traffic is randomly load balanced across IPs in the backend IPs.
        - ``allocated_outbound_ports``: The number of outbound ports to be used for NAT.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure load balancer exists:
            azurearm_network.load_balancer_present:
                - name: lb1
                - resource_group: group1
                - location: eastus
                - frontend_ip_configurations:
                  - name: lb1_feip1
                    public_ip_address: pub_ip1
                - backend_address_pools:
                  - name: lb1_bepool1
                - probes:
                  - name: lb1_webprobe1
                    protocol: tcp
                    port: 80
                    interval_in_seconds: 5
                    number_of_probes: 2
                - load_balancing_rules:
                  - name: lb1_webprobe1
                    protocol: tcp
                    frontend_port: 80
                    backend_port: 80
                    idle_timeout_in_minutes: 4
                    frontend_ip_configuration: lb1_feip1
                    backend_address_pool: lb1_bepool1
                    probe: lb1_webprobe1
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_resource: Ensure resource group exists
                  - azurearm_network: Ensure public IP exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    if sku:
        sku = {"name": sku.capitalize()}

    load_bal = __salt__["azurearm_network.load_balancer_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in load_bal:
        # tag changes
        tag_changes = __utils__["dictdiffer.deep_diff"](
            load_bal.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        # sku changes
        if sku:
            sku_changes = __utils__["dictdiffer.deep_diff"](
                load_bal.get("sku", {}), sku
            )
            if sku_changes:
                ret["changes"]["sku"] = sku_changes

        # frontend_ip_configurations changes
        if frontend_ip_configurations:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("frontend_ip_configurations", []),
                frontend_ip_configurations,
                ["public_ip_address", "subnet"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"frontend_ip_configurations" {}'.format(
                    comp_ret["comment"]
                )
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["frontend_ip_configurations"] = comp_ret["changes"]

        # backend_address_pools changes
        if backend_address_pools:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("backend_address_pools", []), backend_address_pools
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"backend_address_pools" {}'.format(
                    comp_ret["comment"]
                )
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["backend_address_pools"] = comp_ret["changes"]

        # probes changes
        if probes:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("probes", []), probes
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"probes" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["probes"] = comp_ret["changes"]

        # load_balancing_rules changes
        if load_balancing_rules:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("load_balancing_rules", []),
                load_balancing_rules,
                ["frontend_ip_configuration", "backend_address_pool", "probe"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"load_balancing_rules" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["load_balancing_rules"] = comp_ret["changes"]

        # inbound_nat_rules changes
        if inbound_nat_rules:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("inbound_nat_rules", []),
                inbound_nat_rules,
                ["frontend_ip_configuration"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"inbound_nat_rules" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["inbound_nat_rules"] = comp_ret["changes"]

        # inbound_nat_pools changes
        if inbound_nat_pools:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("inbound_nat_pools", []),
                inbound_nat_pools,
                ["frontend_ip_configuration"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"inbound_nat_pools" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["inbound_nat_pools"] = comp_ret["changes"]

        # outbound_nat_rules changes
        if outbound_nat_rules:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                load_bal.get("outbound_nat_rules", []),
                outbound_nat_rules,
                ["frontend_ip_configuration"],
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"outbound_nat_rules" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["outbound_nat_rules"] = comp_ret["changes"]

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Load balancer {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Load balancer {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "sku": sku,
                "tags": tags,
                "frontend_ip_configurations": frontend_ip_configurations,
                "backend_address_pools": backend_address_pools,
                "load_balancing_rules": load_balancing_rules,
                "probes": probes,
                "inbound_nat_rules": inbound_nat_rules,
                "inbound_nat_pools": inbound_nat_pools,
                "outbound_nat_rules": outbound_nat_rules,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Load balancer {name} would be created."
        ret["result"] = None
        return ret

    lb_kwargs = kwargs.copy()
    lb_kwargs.update(connection_auth)

    load_bal = __salt__["azurearm_network.load_balancer_create_or_update"](
        name=name,
        resource_group=resource_group,
        sku=sku,
        tags=tags,
        frontend_ip_configurations=frontend_ip_configurations,
        backend_address_pools=backend_address_pools,
        load_balancing_rules=load_balancing_rules,
        probes=probes,
        inbound_nat_rules=inbound_nat_rules,
        inbound_nat_pools=inbound_nat_pools,
        outbound_nat_rules=outbound_nat_rules,
        **lb_kwargs,
    )

    if "error" not in load_bal:
        ret["result"] = True
        ret["comment"] = f"Load balancer {name} has been created."
        return ret

    ret["comment"] = "Failed to create load balancer {}! ({})".format(
        name, load_bal.get("error")
    )
    return ret


@_deprecation_message
def load_balancer_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a load balancer does not exist in the resource group.

    :param name:
        Name of the load balancer.

    :param resource_group:
        The resource group assigned to the load balancer.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    load_bal = __salt__["azurearm_network.load_balancer_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in load_bal:
        ret["result"] = True
        ret["comment"] = f"Load balancer {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Load balancer {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": load_bal,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.load_balancer_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Load balancer {name} has been deleted."
        ret["changes"] = {"old": load_bal, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete load balancer {name}!"
    return ret


@_deprecation_message
def public_ip_address_present(
    name,
    resource_group,
    tags=None,
    sku=None,
    public_ip_allocation_method=None,
    public_ip_address_version=None,
    dns_settings=None,
    idle_timeout_in_minutes=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a public IP address exists.

    :param name:
        Name of the public IP address.

    :param resource_group:
        The resource group assigned to the public IP address.

    :param dns_settings:
        An optional dictionary representing a valid PublicIPAddressDnsSettings object. Parameters include
        'domain_name_label' and 'reverse_fqdn', which accept strings. The 'domain_name_label' parameter is concatenated
        with the regionalized DNS zone make up the fully qualified domain name associated with the public IP address.
        If a domain name label is specified, an A DNS record is created for the public IP in the Microsoft Azure DNS
        system. The 'reverse_fqdn' parameter is a user-visible, fully qualified domain name that resolves to this public
        IP address. If the reverse FQDN is specified, then a PTR DNS record is created pointing from the IP address in
        the in-addr.arpa domain to the reverse FQDN.

    :param sku:
        The public IP address SKU, which can be 'Basic' or 'Standard'.

    :param public_ip_allocation_method:
        The public IP allocation method. Possible values are: 'Static' and 'Dynamic'.

    :param public_ip_address_version:
        The public IP address version. Possible values are: 'IPv4' and 'IPv6'.

    :param idle_timeout_in_minutes:
        An integer representing the idle timeout of the public IP address.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the public IP address object.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure public IP exists:
            azurearm_network.public_ip_address_present:
                - name: pub_ip1
                - resource_group: group1
                - dns_settings:
                    domain_name_label: decisionlab-ext-test-label
                - sku: basic
                - public_ip_allocation_method: static
                - public_ip_address_version: ipv4
                - idle_timeout_in_minutes: 4
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_resource: Ensure resource group exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    if sku:
        sku = {"name": sku.capitalize()}

    pub_ip = __salt__["azurearm_network.public_ip_address_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in pub_ip:
        # tag changes
        tag_changes = __utils__["dictdiffer.deep_diff"](
            pub_ip.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        # dns_settings changes
        if dns_settings:
            if not isinstance(dns_settings, dict):
                ret["comment"] = "DNS settings must be provided as a dictionary!"
                return ret

            for key in dns_settings:
                if dns_settings[key] != pub_ip.get("dns_settings", {}).get(key):
                    ret["changes"]["dns_settings"] = {
                        "old": pub_ip.get("dns_settings"),
                        "new": dns_settings,
                    }
                    break

        # sku changes
        if sku:
            sku_changes = __utils__["dictdiffer.deep_diff"](pub_ip.get("sku", {}), sku)
            if sku_changes:
                ret["changes"]["sku"] = sku_changes

        # public_ip_allocation_method changes
        if public_ip_allocation_method:
            if public_ip_allocation_method.capitalize() != pub_ip.get(
                "public_ip_allocation_method"
            ):
                ret["changes"]["public_ip_allocation_method"] = {
                    "old": pub_ip.get("public_ip_allocation_method"),
                    "new": public_ip_allocation_method,
                }

        # public_ip_address_version changes
        if public_ip_address_version:
            if (
                public_ip_address_version.lower()
                != pub_ip.get("public_ip_address_version", "").lower()
            ):
                ret["changes"]["public_ip_address_version"] = {
                    "old": pub_ip.get("public_ip_address_version"),
                    "new": public_ip_address_version,
                }

        # idle_timeout_in_minutes changes
        if idle_timeout_in_minutes and (
            int(idle_timeout_in_minutes) != pub_ip.get("idle_timeout_in_minutes")
        ):
            ret["changes"]["idle_timeout_in_minutes"] = {
                "old": pub_ip.get("idle_timeout_in_minutes"),
                "new": idle_timeout_in_minutes,
            }

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Public IP address {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Public IP address {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "tags": tags,
                "dns_settings": dns_settings,
                "sku": sku,
                "public_ip_allocation_method": public_ip_allocation_method,
                "public_ip_address_version": public_ip_address_version,
                "idle_timeout_in_minutes": idle_timeout_in_minutes,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Public IP address {name} would be created."
        ret["result"] = None
        return ret

    pub_ip_kwargs = kwargs.copy()
    pub_ip_kwargs.update(connection_auth)

    pub_ip = __salt__["azurearm_network.public_ip_address_create_or_update"](
        name=name,
        resource_group=resource_group,
        sku=sku,
        tags=tags,
        dns_settings=dns_settings,
        public_ip_allocation_method=public_ip_allocation_method,
        public_ip_address_version=public_ip_address_version,
        idle_timeout_in_minutes=idle_timeout_in_minutes,
        **pub_ip_kwargs,
    )

    if "error" not in pub_ip:
        ret["result"] = True
        ret["comment"] = f"Public IP address {name} has been created."
        return ret

    ret["comment"] = "Failed to create public IP address {}! ({})".format(
        name, pub_ip.get("error")
    )
    return ret


@_deprecation_message
def public_ip_address_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a public IP address does not exist in the resource group.

    :param name:
        Name of the public IP address.

    :param resource_group:
        The resource group assigned to the public IP address.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    pub_ip = __salt__["azurearm_network.public_ip_address_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in pub_ip:
        ret["result"] = True
        ret["comment"] = f"Public IP address {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Public IP address {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": pub_ip,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.public_ip_address_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Public IP address {name} has been deleted."
        ret["changes"] = {"old": pub_ip, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete public IP address {name}!"
    return ret


@_deprecation_message
def network_interface_present(
    name,
    ip_configurations,
    subnet,
    virtual_network,
    resource_group,
    tags=None,
    virtual_machine=None,
    network_security_group=None,
    dns_settings=None,
    mac_address=None,
    primary=None,
    enable_accelerated_networking=None,
    enable_ip_forwarding=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a network interface exists.

    :param name:
        Name of the network interface.

    :param ip_configurations:
        A list of dictionaries representing valid NetworkInterfaceIPConfiguration objects. The 'name' key is required at
        minimum. At least one IP Configuration must be present.

    :param subnet:
        Name of the existing subnet assigned to the network interface.

    :param virtual_network:
        Name of the existing virtual network containing the subnet.

    :param resource_group:
        The resource group assigned to the virtual network.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the network interface object.

    :param network_security_group:
        The name of the existing network security group to assign to the network interface.

    :param virtual_machine:
        The name of the existing virtual machine to assign to the network interface.

    :param dns_settings:
        An optional dictionary representing a valid NetworkInterfaceDnsSettings object. Valid parameters are:

        - ``dns_servers``: List of DNS server IP addresses. Use 'AzureProvidedDNS' to switch to Azure provided DNS
          resolution. 'AzureProvidedDNS' value cannot be combined with other IPs, it must be the only value in
          dns_servers collection.
        - ``internal_dns_name_label``: Relative DNS name for this NIC used for internal communications between VMs in
          the same virtual network.
        - ``internal_fqdn``: Fully qualified DNS name supporting internal communications between VMs in the same virtual
          network.
        - ``internal_domain_name_suffix``: Even if internal_dns_name_label is not specified, a DNS entry is created for
          the primary NIC of the VM. This DNS name can be constructed by concatenating the VM name with the value of
          internal_domain_name_suffix.

    :param mac_address:
        Optional string containing the MAC address of the network interface.

    :param primary:
        Optional boolean allowing the interface to be set as the primary network interface on a virtual machine
        with multiple interfaces attached.

    :param enable_accelerated_networking:
        Optional boolean indicating whether accelerated networking should be enabled for the interface.

    :param enable_ip_forwarding:
        Optional boolean indicating whether IP forwarding should be enabled for the interface.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure network interface exists:
            azurearm_network.network_interface_present:
                - name: iface1
                - subnet: vnet1_sn1
                - virtual_network: vnet1
                - resource_group: group1
                - ip_configurations:
                  - name: iface1_ipc1
                    public_ip_address: pub_ip2
                - dns_settings:
                    internal_dns_name_label: decisionlab-int-test-label
                - primary: True
                - enable_accelerated_networking: True
                - enable_ip_forwarding: False
                - network_security_group: nsg1
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_network: Ensure subnet exists
                  - azurearm_network: Ensure network security group exists
                  - azurearm_network: Ensure another public IP exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    iface = __salt__["azurearm_network.network_interface_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in iface:
        # tag changes
        tag_changes = __utils__["dictdiffer.deep_diff"](
            iface.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        # mac_address changes
        if mac_address and (mac_address != iface.get("mac_address")):
            ret["changes"]["mac_address"] = {
                "old": iface.get("mac_address"),
                "new": mac_address,
            }

        # primary changes
        if primary is not None:
            if primary != iface.get("primary", True):
                ret["changes"]["primary"] = {
                    "old": iface.get("primary"),
                    "new": primary,
                }

        # enable_accelerated_networking changes
        if enable_accelerated_networking is not None:
            if enable_accelerated_networking != iface.get(
                "enable_accelerated_networking"
            ):
                ret["changes"]["enable_accelerated_networking"] = {
                    "old": iface.get("enable_accelerated_networking"),
                    "new": enable_accelerated_networking,
                }

        # enable_ip_forwarding changes
        if enable_ip_forwarding is not None:
            if enable_ip_forwarding != iface.get("enable_ip_forwarding"):
                ret["changes"]["enable_ip_forwarding"] = {
                    "old": iface.get("enable_ip_forwarding"),
                    "new": enable_ip_forwarding,
                }

        # network_security_group changes
        nsg_name = None
        if iface.get("network_security_group"):
            nsg_name = iface["network_security_group"]["id"].split("/")[-1]

        if network_security_group and (network_security_group != nsg_name):
            ret["changes"]["network_security_group"] = {
                "old": nsg_name,
                "new": network_security_group,
            }

        # virtual_machine changes
        vm_name = None
        if iface.get("virtual_machine"):
            vm_name = iface["virtual_machine"]["id"].split("/")[-1]

        if virtual_machine and (virtual_machine != vm_name):
            ret["changes"]["virtual_machine"] = {"old": vm_name, "new": virtual_machine}

        # dns_settings changes
        if dns_settings:
            if not isinstance(dns_settings, dict):
                ret["comment"] = "DNS settings must be provided as a dictionary!"
                return ret

            for key in dns_settings:
                if (
                    dns_settings[key].lower()
                    != iface.get("dns_settings", {}).get(key, "").lower()
                ):
                    ret["changes"]["dns_settings"] = {
                        "old": iface.get("dns_settings"),
                        "new": dns_settings,
                    }
                    break

        # ip_configurations changes
        comp_ret = __utils__["azurearm.compare_list_of_dicts"](
            iface.get("ip_configurations", []),
            ip_configurations,
            ["public_ip_address", "subnet"],
        )

        if comp_ret.get("comment"):
            ret["comment"] = '"ip_configurations" {}'.format(comp_ret["comment"])
            return ret

        if comp_ret.get("changes"):
            ret["changes"]["ip_configurations"] = comp_ret["changes"]

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Network interface {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Network interface {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "ip_configurations": ip_configurations,
                "dns_settings": dns_settings,
                "network_security_group": network_security_group,
                "virtual_machine": virtual_machine,
                "enable_accelerated_networking": enable_accelerated_networking,
                "enable_ip_forwarding": enable_ip_forwarding,
                "mac_address": mac_address,
                "primary": primary,
                "tags": tags,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Network interface {name} would be created."
        ret["result"] = None
        return ret

    iface_kwargs = kwargs.copy()
    iface_kwargs.update(connection_auth)

    iface = __salt__["azurearm_network.network_interface_create_or_update"](
        name=name,
        subnet=subnet,
        virtual_network=virtual_network,
        resource_group=resource_group,
        ip_configurations=ip_configurations,
        dns_settings=dns_settings,
        enable_accelerated_networking=enable_accelerated_networking,
        enable_ip_forwarding=enable_ip_forwarding,
        mac_address=mac_address,
        primary=primary,
        network_security_group=network_security_group,
        virtual_machine=virtual_machine,
        tags=tags,
        **iface_kwargs,
    )

    if "error" not in iface:
        ret["result"] = True
        ret["comment"] = f"Network interface {name} has been created."
        return ret

    ret["comment"] = "Failed to create network interface {}! ({})".format(
        name, iface.get("error")
    )
    return ret


@_deprecation_message
def network_interface_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a network interface does not exist in the resource group.

    :param name:
        Name of the network interface.

    :param resource_group:
        The resource group assigned to the network interface.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    iface = __salt__["azurearm_network.network_interface_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in iface:
        ret["result"] = True
        ret["comment"] = f"Network interface {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Network interface {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": iface,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.network_interface_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Network interface {name} has been deleted."
        ret["changes"] = {"old": iface, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete network interface {name}!)"
    return ret


@_deprecation_message
def route_table_present(
    name,
    resource_group,
    tags=None,
    routes=None,
    disable_bgp_route_propagation=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a route table exists.

    :param name:
        Name of the route table.

    :param resource_group:
        The resource group assigned to the route table.

    :param routes:
        An optional list of dictionaries representing valid Route objects contained within a route table. See the
        documentation for the route_present state or route_create_or_update execution module for more information on
        required and optional parameters for routes. The routes are only managed if this parameter is present. When this
        parameter is absent, implemented routes will not be removed, and will merely become unmanaged.

    :param disable_bgp_route_propagation:
        An optional boolean parameter setting whether to disable the routes learned by BGP on the route table.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the route table object.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure route table exists:
            azurearm_network.route_table_present:
                - name: rt1
                - resource_group: group1
                - routes:
                  - name: rt1_route1
                    address_prefix: '0.0.0.0/0'
                    next_hop_type: internet
                  - name: rt1_route2
                    address_prefix: '192.168.0.0/16'
                    next_hop_type: vnetlocal
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_resource: Ensure resource group exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    rt_tbl = __salt__["azurearm_network.route_table_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in rt_tbl:
        # tag changes
        tag_changes = __utils__["dictdiffer.deep_diff"](
            rt_tbl.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        # disable_bgp_route_propagation changes
        # pylint: disable=line-too-long
        if disable_bgp_route_propagation and (
            disable_bgp_route_propagation != rt_tbl.get("disable_bgp_route_propagation")
        ):
            ret["changes"]["disable_bgp_route_propagation"] = {
                "old": rt_tbl.get("disable_bgp_route_propagation"),
                "new": disable_bgp_route_propagation,
            }

        # routes changes
        if routes:
            comp_ret = __utils__["azurearm.compare_list_of_dicts"](
                rt_tbl.get("routes", []), routes
            )

            if comp_ret.get("comment"):
                ret["comment"] = '"routes" {}'.format(comp_ret["comment"])
                return ret

            if comp_ret.get("changes"):
                ret["changes"]["routes"] = comp_ret["changes"]

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Route table {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Route table {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "tags": tags,
                "routes": routes,
                "disable_bgp_route_propagation": disable_bgp_route_propagation,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Route table {name} would be created."
        ret["result"] = None
        return ret

    rt_tbl_kwargs = kwargs.copy()
    rt_tbl_kwargs.update(connection_auth)

    rt_tbl = __salt__["azurearm_network.route_table_create_or_update"](
        name=name,
        resource_group=resource_group,
        disable_bgp_route_propagation=disable_bgp_route_propagation,
        routes=routes,
        tags=tags,
        **rt_tbl_kwargs,
    )

    if "error" not in rt_tbl:
        ret["result"] = True
        ret["comment"] = f"Route table {name} has been created."
        return ret

    ret["comment"] = "Failed to create route table {}! ({})".format(
        name, rt_tbl.get("error")
    )
    return ret


@_deprecation_message
def route_table_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a route table does not exist in the resource group.

    :param name:
        Name of the route table.

    :param resource_group:
        The resource group assigned to the route table.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    rt_tbl = __salt__["azurearm_network.route_table_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in rt_tbl:
        ret["result"] = True
        ret["comment"] = f"Route table {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Route table {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": rt_tbl,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.route_table_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Route table {name} has been deleted."
        ret["changes"] = {"old": rt_tbl, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete route table {name}!"
    return ret


@_deprecation_message
def route_present(
    name,
    address_prefix,
    next_hop_type,
    route_table,
    resource_group,
    next_hop_ip_address=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure a route exists within a route table.

    :param name:
        Name of the route.

    :param address_prefix:
        The destination CIDR to which the route applies.

    :param next_hop_type:
        The type of Azure hop the packet should be sent to. Possible values are: 'VirtualNetworkGateway', 'VnetLocal',
        'Internet', 'VirtualAppliance', and 'None'.

    :param next_hop_ip_address:
        The IP address packets should be forwarded to. Next hop values are only allowed in routes where the next hop
        type is 'VirtualAppliance'.

    :param route_table:
        The name of the existing route table which will contain the route.

    :param resource_group:
        The resource group assigned to the route table.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure route exists:
            azurearm_network.route_present:
                - name: rt1_route2
                - route_table: rt1
                - resource_group: group1
                - address_prefix: '192.168.0.0/16'
                - next_hop_type: vnetlocal
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_network: Ensure route table exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    route = __salt__["azurearm_network.route_get"](
        name, route_table, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in route:
        if address_prefix != route.get("address_prefix"):
            ret["changes"]["address_prefix"] = {
                "old": route.get("address_prefix"),
                "new": address_prefix,
            }

        if next_hop_type.lower() != route.get("next_hop_type", "").lower():
            ret["changes"]["next_hop_type"] = {
                "old": route.get("next_hop_type"),
                "new": next_hop_type,
            }

        if (
            next_hop_type.lower() == "virtualappliance"
            and next_hop_ip_address != route.get("next_hop_ip_address")
        ):
            ret["changes"]["next_hop_ip_address"] = {
                "old": route.get("next_hop_ip_address"),
                "new": next_hop_ip_address,
            }

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Route {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Route {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "address_prefix": address_prefix,
                "next_hop_type": next_hop_type,
                "next_hop_ip_address": next_hop_ip_address,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Route {name} would be created."
        ret["result"] = None
        return ret

    route_kwargs = kwargs.copy()
    route_kwargs.update(connection_auth)

    route = __salt__["azurearm_network.route_create_or_update"](
        name=name,
        route_table=route_table,
        resource_group=resource_group,
        address_prefix=address_prefix,
        next_hop_type=next_hop_type,
        next_hop_ip_address=next_hop_ip_address,
        **route_kwargs,
    )

    if "error" not in route:
        ret["result"] = True
        ret["comment"] = f"Route {name} has been created."
        return ret

    ret["comment"] = "Failed to create route {}! ({})".format(name, route.get("error"))
    return ret


@_deprecation_message
def route_absent(name, route_table, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a route table does not exist in the resource group.

    :param name:
        Name of the route table.

    :param route_table:
        The name of the existing route table containing the route.

    :param resource_group:
        The resource group assigned to the route table.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    route = __salt__["azurearm_network.route_get"](
        name, route_table, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in route:
        ret["result"] = True
        ret["comment"] = f"Route {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Route {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": route,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_network.route_delete"](
        name, route_table, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Route {name} has been deleted."
        ret["changes"] = {"old": route, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete route {name}!"
    return ret



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import logging

import pytest

import salt.exceptions
import salt.state
import salt.utils.files
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.core_test,
]


def test_format_log_non_ascii_character():
    """
    Tests running a non-ascii character through the state.format_log
    function. See Issue #33605.
    """
    # There is no return to test against as the format_log
    # function doesn't return anything. However, we do want
    # to make sure that the function doesn't stacktrace when
    # called.
    ret = {
        "changes": {"Français": {"old": "something old", "new": "something new"}},
        "result": True,
    }
    salt.state.format_log(ret)


def test_format_log_list(caplog):
    """
    Test running format_log when ret is not a dictionary
    """
    with caplog.at_level(logging.INFO):
        ret = ["test1", "test2"]
        salt.state.format_log(ret)
        assert "INFO" in caplog.text
        assert f"{ret}" in caplog.text


def test_render_error_on_invalid_requisite(minion_opts):
    """
    Test that the state compiler correctly deliver a rendering
    exception when a requisite cannot be resolved
    """
    with patch("salt.state.State._gather_pillar"):
        high_data = {
            "git": salt.state.HashableOrderedDict(
                [
                    (
                        "pkg",
                        [
                            salt.state.HashableOrderedDict(
                                [
                                    (
                                        "require",
                                        [
                                            salt.state.HashableOrderedDict(
                                                [
                                                    (
                                                        "file",
                                                        salt.state.HashableOrderedDict(
                                                            [("test1", "test")]
                                                        ),
                                                    )
                                                ]
                                            )
                                        ],
                                    )
                                ]
                            ),
                            "installed",
                            {"order": 10000},
                        ],
                    ),
                    ("__sls__", "issue_35226"),
                    ("__env__", "base"),
                ]
            )
        }
        minion_opts["pillar"] = {
            "git": salt.state.HashableOrderedDict([("test1", "test")])
        }
        state_obj = salt.state.State(minion_opts)
        with pytest.raises(salt.exceptions.SaltRenderError):
            state_obj.call_high(high_data)


def test_verify_onlyif_parse(minion_opts):
    low_data = {
        "onlyif": [{"fun": "test.arg", "args": ["arg1", "arg2"]}],
        "name": "mysql-server-5.7",
        "state": "debconf",
        "__id__": "set root password",
        "fun": "set",
        "__env__": "base",
        "__sls__": "debconf",
        "data": {
            "mysql-server/root_password": {"type": "password", "value": "temp123"}
        },
        "order": 10000,
    }
    expected_result = {"comment": "onlyif condition is true", "result": False}
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, "")
        assert expected_result == return_result


def test_verify_onlyif_parse_deep_return(minion_opts):
    low_data = {
        "state": "test",
        "name": "foo",
        "__sls__": "consol",
        "__env__": "base",
        "__id__": "test",
        "onlyif": [
            {
                "fun": "test.arg",
                "get_return": "kwargs:deep:return",
                "deep": {"return": "true"},
            }
        ],
        "order": 10000,
        "fun": "nop",
    }
    expected_result = {"comment": "onlyif condition is true", "result": False}
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, "")
        assert expected_result == return_result


def test_verify_onlyif_cmd_error(minion_opts):
    """
    Simulates a failure in cmd.retcode from onlyif
    This could occur if runas is specified with a user that does not exist
    """
    low_data = {
        "onlyif": "somecommand",
        "runas": "doesntexist",
        "name": "echo something",
        "state": "cmd",
        "__id__": "this is just a test",
        "fun": "run",
        "__env__": "base",
        "__sls__": "sometest",
        "order": 10000,
    }
    expected_result = {
        "comment": "onlyif condition is false",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        mock = MagicMock(side_effect=CommandExecutionError("Boom!"))
        with patch.dict(state_obj.functions, {"cmd.retcode": mock}):
            #  The mock handles the exception, but the runas dict is being passed as it would actually be
            return_result = state_obj._run_check_onlyif(
                low_data, {"runas": "doesntexist"}
            )
            assert expected_result == return_result


def test_verify_unless_cmd_error(minion_opts):
    """
    Simulates a failure in cmd.retcode from unless
    This could occur if runas is specified with a user that does not exist
    """
    low_data = {
        "unless": "somecommand",
        "runas": "doesntexist",
        "name": "echo something",
        "state": "cmd",
        "__id__": "this is just a test",
        "fun": "run",
        "__env__": "base",
        "__sls__": "sometest",
        "order": 10000,
    }
    expected_result = {
        "comment": "unless condition is true",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        mock = MagicMock(side_effect=CommandExecutionError("Boom!"))
        with patch.dict(state_obj.functions, {"cmd.retcode": mock}):
            #  The mock handles the exception, but the runas dict is being passed as it would actually be
            return_result = state_obj._run_check_unless(
                low_data, {"runas": "doesntexist"}
            )
            assert expected_result == return_result


def test_verify_unless_list_cmd(minion_opts):
    """
    If any of the unless commands return False (non 0) then the state should
    run (no skip_watch).
    """
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check unless",
        "unless": ["exit 0", "exit 1"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {
        "comment": "unless condition is false",
        "result": False,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, {})
        assert expected_result == return_result


def test_verify_unless_list_cmd_different_order(minion_opts):
    """
    If any of the unless commands return False (non 0) then the state should
    run (no skip_watch). The order shouldn't matter.
    """
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check unless",
        "unless": ["exit 1", "exit 0"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {
        "comment": "unless condition is false",
        "result": False,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, {})
        assert expected_result == return_result


def test_verify_onlyif_list_cmd_different_order(minion_opts):
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check onlyif",
        "onlyif": ["exit 1", "exit 0"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {
        "comment": "onlyif condition is false",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, {})
        assert expected_result == return_result


def test_verify_unless_list_cmd_valid(minion_opts):
    """
    If any of the unless commands return False (non 0) then the state should
    run (no skip_watch). This tests all commands return False.
    """
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check unless",
        "unless": ["exit 1", "exit 1"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {"comment": "unless condition is false", "result": False}
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, {})
        assert expected_result == return_result


def test_verify_onlyif_list_cmd_valid(minion_opts):
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check onlyif",
        "onlyif": ["exit 0", "exit 0"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {"comment": "onlyif condition is true", "result": False}
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, {})
        assert expected_result == return_result


def test_verify_unless_list_cmd_invalid(minion_opts):
    """
    If any of the unless commands return False (non 0) then the state should
    run (no skip_watch). This tests all commands return True
    """
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check unless",
        "unless": ["exit 0", "exit 0"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {
        "comment": "unless condition is true",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, {})
        assert expected_result == return_result


def test_verify_onlyif_list_cmd_invalid(minion_opts):
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check onlyif",
        "onlyif": ["exit 1", "exit 1"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {
        "comment": "onlyif condition is false",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, {})
        assert expected_result == return_result


def test_verify_unless_parse(minion_opts):
    low_data = {
        "unless": [{"fun": "test.arg", "args": ["arg1", "arg2"]}],
        "name": "mysql-server-5.7",
        "state": "debconf",
        "__id__": "set root password",
        "fun": "set",
        "__env__": "base",
        "__sls__": "debconf",
        "data": {
            "mysql-server/root_password": {"type": "password", "value": "temp123"}
        },
        "order": 10000,
    }
    expected_result = {
        "comment": "unless condition is true",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, "")
        assert expected_result == return_result


def test_verify_unless_parse_deep_return(minion_opts):
    low_data = {
        "state": "test",
        "name": "foo",
        "__sls__": "consol",
        "__env__": "base",
        "__id__": "test",
        "unless": [
            {
                "fun": "test.arg",
                "get_return": "kwargs:deep:return",
                "deep": {"return": False},
            }
        ],
        "order": 10000,
        "fun": "nop",
    }
    expected_result = {"comment": "unless condition is false", "result": False}
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, "")
        assert expected_result == return_result


def test_verify_creates(minion_opts):
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.creates",
        "__env__": "base",
        "__id__": "do_a_thing",
        "creates": "/tmp/thing",
        "order": 10000,
        "fun": "run",
    }

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        with patch("os.path.exists") as path_mock:
            path_mock.return_value = True
            expected_result = {
                "comment": "/tmp/thing exists",
                "result": True,
                "skip_watch": True,
            }
            for key in ("__sls__", "__id__", "name"):
                expected_result[key] = low_data.get(key)
            return_result = state_obj._run_check_creates(low_data)
            assert expected_result == return_result

            path_mock.return_value = False
            expected_result = {
                "comment": "Creates files not found",
                "result": False,
            }
            for key in ("__sls__", "__id__", "name"):
                expected_result[key] = low_data.get(key)
            return_result = state_obj._run_check_creates(low_data)
            assert expected_result == return_result


def test_verify_creates_list(minion_opts):
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.creates",
        "__env__": "base",
        "__id__": "do_a_thing",
        "creates": ["/tmp/thing", "/tmp/thing2"],
        "order": 10000,
        "fun": "run",
    }

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        with patch("os.path.exists") as path_mock:
            path_mock.return_value = True
            expected_result = {
                "comment": "All files in creates exist",
                "result": True,
                "skip_watch": True,
            }
            for key in ("__sls__", "__id__", "name"):
                expected_result[key] = low_data.get(key)
            return_result = state_obj._run_check_creates(low_data)
            assert expected_result == return_result

            path_mock.return_value = False
            expected_result = {
                "comment": "Creates files not found",
                "result": False,
            }
            for key in ("__sls__", "__id__", "name"):
                expected_result[key] = low_data.get(key)
            return_result = state_obj._run_check_creates(low_data)
            assert expected_result == return_result


def _expand_win_path(path):
    """
    Expand C:/users/admini~1/appdata/local/temp/salt-tests-tmpdir/...
    into C:/users/adminitrator/appdata/local/temp/salt-tests-tmpdir/...
    to prevent file.search from expanding the "~" with os.path.expanduser
    """
    if salt.utils.platform.is_windows():
        import win32file

        return win32file.GetLongPathName(path).replace("\\", "/")
    else:
        return path


def test_verify_onlyif_parse_slots(tmp_path, minion_opts):
    name = str(tmp_path / "testfile.txt")
    with salt.utils.files.fopen(name, "w") as fp:
        fp.write("file-contents")
    low_data = {
        "onlyif": [
            {
                "fun": "file.search",
                "args": [f"__slot__:salt:test.echo({_expand_win_path(name)})"],
                "pattern": "__slot__:salt:test.echo(file-contents)",
            }
        ],
        "name": "mysql-server-5.7",
        "state": "debconf",
        "__id__": "set root password",
        "fun": "set",
        "__env__": "base",
        "__sls__": "debconf",
        "data": {
            "mysql-server/root_password": {"type": "password", "value": "temp123"}
        },
        "order": 10000,
    }
    expected_result = {"comment": "onlyif condition is true", "result": False}
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, "")
        assert expected_result == return_result


def test_verify_onlyif_list_cmd(minion_opts):
    low_data = {
        "state": "cmd",
        "name": 'echo "something"',
        "__sls__": "tests.cmd",
        "__env__": "base",
        "__id__": "check onlyif",
        "onlyif": ["exit 0", "exit 1"],
        "order": 10001,
        "fun": "run",
    }
    expected_result = {
        "comment": "onlyif condition is false",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_onlyif(low_data, {})
        assert expected_result == return_result


def test_verify_onlyif_cmd_args(minion_opts):
    """
    Verify cmd.run state arguments are properly passed to cmd.retcode in onlyif
    """
    low_data = {
        "onlyif": "somecommand",
        "cwd": "acwd",
        "root": "aroot",
        "env": [{"akey": "avalue"}],
        "prepend_path": "apath",
        "umask": "0700",
        "success_retcodes": 1,
        "timeout": 5,
        "runas": "doesntexist",
        "name": "echo something",
        "shell": "/bin/dash",
        "state": "cmd",
        "__id__": "this is just a test",
        "fun": "run",
        "__env__": "base",
        "__sls__": "sometest",
        "order": 10000,
    }

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        mock = MagicMock()
        with patch.dict(state_obj.functions, {"cmd.retcode": mock}):
            #  The mock handles the exception, but the runas dict is being passed as it would actually be
            return_result = state_obj._run_check(low_data)
            mock.assert_called_once_with(
                "somecommand",
                ignore_retcode=True,
                python_shell=True,
                cwd="acwd",
                root="aroot",
                runas="doesntexist",
                env=[{"akey": "avalue"}],
                prepend_path="apath",
                umask="0700",
                timeout=5,
                success_retcodes=1,
                shell="/bin/dash",
            )


def test_verify_unless_parse_slots(tmp_path, minion_opts):
    name = str(tmp_path / "testfile.txt")
    with salt.utils.files.fopen(name, "w") as fp:
        fp.write("file-contents")
    low_data = {
        "unless": [
            {
                "fun": "file.search",
                "args": [f"__slot__:salt:test.echo({_expand_win_path(name)})"],
                "pattern": "__slot__:salt:test.echo(file-contents)",
            }
        ],
        "name": "mysql-server-5.7",
        "state": "debconf",
        "__id__": "set root password",
        "fun": "set",
        "__env__": "base",
        "__sls__": "debconf",
        "data": {
            "mysql-server/root_password": {"type": "password", "value": "temp123"}
        },
        "order": 10000,
    }
    expected_result = {
        "comment": "unless condition is true",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        return_result = state_obj._run_check_unless(low_data, "")
        assert expected_result == return_result


def test_verify_retry_parsing(minion_opts):
    low_data = {
        "state": "file",
        "name": "/tmp/saltstack.README.rst",
        "__sls__": "demo.download",
        "__env__": "base",
        "__id__": "download sample data",
        "retry": {"attempts": 5, "interval": 5},
        "unless": ["test -f /tmp/saltstack.README.rst"],
        "source": [
            "https://raw.githubusercontent.com/saltstack/salt/develop/README.rst"
        ],
        "source_hash": "f2bc8c0aa2ae4f5bb5c2051686016b48",
        "order": 10000,
        "fun": "managed",
    }
    expected_result = {
        "__run_num__": 0,
        "changes": {},
        "comment": "['unless condition is true']  The state would be retried every 5 "
        "seconds (with a splay of up to 0 seconds) a maximum of 5 times or "
        "until a result of True is returned",
        "result": True,
        "skip_watch": True,
    }
    for key in ("__sls__", "__id__", "name"):
        expected_result[key] = low_data.get(key)

    with patch("salt.state.State._gather_pillar"):
        minion_opts["test"] = True
        minion_opts["file_client"] = "local"
        state_obj = salt.state.State(minion_opts)
        mock = {
            "result": True,
            "comment": ["unless condition is true"],
            "skip_watch": True,
        }
        with patch.object(state_obj, "_run_check", return_value=mock):
            assert set(expected_result).issubset(set(state_obj.call(low_data)))


def test_render_requisite_require_disabled(minion_opts):
    """
    Test that the state compiler correctly deliver a rendering
    exception when a requisite cannot be resolved
    """
    with patch("salt.state.State._gather_pillar"):
        high_data = {
            "step_one": salt.state.HashableOrderedDict(
                [
                    (
                        "test",
                        [
                            salt.state.HashableOrderedDict(
                                [
                                    (
                                        "require",
                                        [
                                            salt.state.HashableOrderedDict(
                                                [("test", "step_two")]
                                            )
                                        ],
                                    )
                                ]
                            ),
                            "succeed_with_changes",
                            {"order": 10000},
                        ],
                    ),
                    ("__sls__", "test.disable_require"),
                    ("__env__", "base"),
                ]
            ),
            "step_two": {
                "test": ["succeed_with_changes", {"order": 10001}],
                "__env__": "base",
                "__sls__": "test.disable_require",
            },
        }

        minion_opts["disabled_requisites"] = ["require"]
        state_obj = salt.state.State(minion_opts)
        ret = state_obj.call_high(high_data)
        run_num = ret["test_|-step_one_|-step_one_|-succeed_with_changes"][
            "__run_num__"
        ]
        assert run_num == 0


def test_render_requisite_require_in_disabled(minion_opts):
    """
    Test that the state compiler correctly deliver a rendering
    exception when a requisite cannot be resolved
    """
    with patch("salt.state.State._gather_pillar"):
        high_data = {
            "step_one": {
                "test": ["succeed_with_changes", {"order": 10000}],
                "__env__": "base",
                "__sls__": "test.disable_require_in",
            },
            "step_two": salt.state.HashableOrderedDict(
                [
                    (
                        "test",
                        [
                            salt.state.HashableOrderedDict(
                                [
                                    (
                                        "require_in",
                                        [
                                            salt.state.HashableOrderedDict(
                                                [("test", "step_one")]
                                            )
                                        ],
                                    )
                                ]
                            ),
                            "succeed_with_changes",
                            {"order": 10001},
                        ],
                    ),
                    ("__sls__", "test.disable_require_in"),
                    ("__env__", "base"),
                ]
            ),
        }

        minion_opts["disabled_requisites"] = ["require_in"]
        state_obj = salt.state.State(minion_opts)
        ret = state_obj.call_high(high_data)
        run_num = ret["test_|-step_one_|-step_one_|-succeed_with_changes"][
            "__run_num__"
        ]
        assert run_num == 0


def test_call_chunk_sub_state_run(minion_opts):
    """
    Test running a batch of states with an external runner
    that returns sub_state_run
    """
    low_data = {
        "state": "external",
        "name": "external_state_name",
        "__id__": "do_a_thing",
        "__sls__": "external",
        "order": 10000,
        "fun": "state",
    }
    mock_call_return = {
        "__run_num__": 0,
        "sub_state_run": [
            {
                "changes": {},
                "result": True,
                "comment": "",
                "low": {
                    "name": "external_state_name",
                    "__id__": "external_state_id",
                    "state": "external_state",
                    "fun": "external_function",
                },
            }
        ],
    }
    expected_sub_state_tag = (
        "external_state_|-external_state_id_|-external_state_name_|-external_function"
    )
    with patch("salt.state.State._gather_pillar"):
        with patch("salt.state.State.call", return_value=mock_call_return):
            minion_opts["disabled_requisites"] = ["require"]
            state_obj = salt.state.State(minion_opts)
            ret = state_obj.call_chunk(low_data, {}, {})
            sub_state = ret.get(expected_sub_state_tag)
            assert sub_state
            assert sub_state["__run_num__"] == 1
            assert sub_state["name"] == "external_state_name"
            assert sub_state["__state_ran__"]
            assert sub_state["__sls__"] == "external"


def test_aggregate_requisites(minion_opts):
    """
    Test to ensure that the requisites are included in the aggregated low state.
    """
    # The low that is returned from _mod_aggregrate
    low = {
        "state": "pkg",
        "name": "other_pkgs",
        "__sls__": "47628",
        "__env__": "base",
        "__id__": "other_pkgs",
        "pkgs": ["byobu", "vim", "tmux", "google-cloud-sdk"],
        "aggregate": True,
        "order": 10002,
        "fun": "installed",
        "__agg__": True,
    }

    # Chunks that have been processed through the pkg mod_aggregate function
    chunks = [
        {
            "state": "file",
            "name": "/tmp/install-vim",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "/tmp/install-vim",
            "order": 10000,
            "fun": "managed",
        },
        {
            "state": "file",
            "name": "/tmp/install-tmux",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "/tmp/install-tmux",
            "order": 10001,
            "fun": "managed",
        },
        {
            "state": "pkg",
            "name": "other_pkgs",
            "__sls__": "47628",
            "__env __": "base",
            "__id__": "other_pkgs",
            "pkgs": ["byobu"],
            "aggregate": True,
            "order": 10002,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "bc",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "bc",
            "hold": True,
            "__agg__": True,
            "order": 10003,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "vim",
            "__sls__": "47628",
            "__env__": "base",
            "__agg__": True,
            "__id__": "vim",
            "require": ["/tmp/install-vim"],
            "order": 10004,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "tmux",
            "__sls__": "47628",
            "__env__": "base",
            "__agg__": True,
            "__id__": "tmux",
            "require": ["/tmp/install-tmux"],
            "order": 10005,
            "fun": "installed",
        },
        {
            "state": "pkgrepo",
            "name": "deb https://packages.cloud.google.com/apt cloud-sdk main",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "google-cloud-repo",
            "humanname": "Google Cloud SDK",
            "file": "/etc/apt/sources.list.d/google-cloud-sdk.list",
            "key_url": "https://packages.cloud.google.com/apt/doc/apt-key.gpg",
            "order": 10006,
            "fun": "managed",
        },
        {
            "state": "pkg",
            "name": "google-cloud-sdk",
            "__sls__": "47628",
            "__env__": "base",
            "__agg__": True,
            "__id__": "google-cloud-sdk",
            "require": ["google-cloud-repo"],
            "order": 10007,
            "fun": "installed",
        },
    ]

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        low_ret = state_obj._aggregate_requisites(low, chunks)

        # Ensure the low returned contains require
        assert "require" in low_ret

        # Ensure all the requires from pkg states are in low
        assert low_ret["require"] == [
            "/tmp/install-vim",
            "/tmp/install-tmux",
            "google-cloud-repo",
        ]


def test_mod_aggregate(minion_opts):
    """
    Test to ensure that the requisites are included in the aggregated low state.
    """
    # The low that is returned from _mod_aggregrate
    low = {
        "state": "pkg",
        "name": "sl",
        "__sls__": "test.62439",
        "__env__": "base",
        "__id__": "sl",
        "require_in": [salt.state.HashableOrderedDict([("file", "/tmp/foo")])],
        "order": 10002,
        "aggregate": True,
        "fun": "installed",
    }

    # Chunks that have been processed through the pkg mod_aggregate function
    chunks = [
        {
            "state": "file",
            "name": "/tmp/foo",
            "__sls__": "test.62439",
            "__env__": "base",
            "__id__": "/tmp/foo",
            "content": "This is some content",
            "order": 10000,
            "require": [{"pkg": "sl"}],
            "fun": "managed",
        },
        {
            "state": "pkg",
            "name": "figlet",
            "__sls__": "test.62439",
            "__env__": "base",
            "__id__": "figlet",
            "__agg__": True,
            "require": [salt.state.HashableOrderedDict([("file", "/tmp/foo")])],
            "order": 10001,
            "aggregate": True,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "sl",
            "__sls__": "test.62439",
            "__env__": "base",
            "__id__": "sl",
            "require_in": [salt.state.HashableOrderedDict([("file", "/tmp/foo")])],
            "order": 10002,
            "aggregate": True,
            "fun": "installed",
        },
    ]

    running = {}

    mock_pkg_mod_aggregate = {
        "state": "pkg",
        "name": "sl",
        "__sls__": "test.62439",
        "__env__": "base",
        "__id__": "sl",
        "require_in": [salt.state.HashableOrderedDict([("file", "/tmp/foo")])],
        "order": 10002,
        "fun": "installed",
        "__agg__": True,
        "pkgs": ["figlet", "sl"],
    }

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        with patch.dict(
            state_obj.states,
            {"pkg.mod_aggregate": MagicMock(return_value=mock_pkg_mod_aggregate)},
        ):
            low_ret = state_obj._mod_aggregate(low, running, chunks)

            # Ensure the low returned contains require
            assert "require_in" in low_ret

            # Ensure all the requires from pkg states are in low
            assert low_ret["require_in"] == [
                salt.state.HashableOrderedDict([("file", "/tmp/foo")])
            ]

            # Ensure that the require requisite from the
            # figlet state finds its way into this state
            assert "require" in low_ret

            # Ensure pkgs were aggregated
            assert low_ret["pkgs"] == ["figlet", "sl"]


def test_mod_aggregate_order(minion_opts):
    """
    Test to ensure that the state_aggregate setting correctly aggregates package installations
    while respecting the 'require' requisite to enforce execution order.
    """
    # Setup the chunks based on the provided scenario
    chunks = [
        {
            "state": "pkg",
            "name": "first packages",
            "__id__": "first packages",
            "pkgs": ["drpm"],
            "fun": "installed",
            "order": 1,
            "__env__": "base",
            "__sls__": "base",
        },
        {
            "state": "test",
            "name": "requirement",
            "__id__": "requirement",
            "fun": "nop",
            "order": 2,
            "__env__": "base",
            "__sls__": "base",
        },
        {
            "state": "pkg",
            "name": "second packages",
            "__id__": "second packages",
            "pkgs": ["gc"],
            "fun": "installed",
            "order": 3,
            "require": [{"test": "requirement"}],
            "provider": "yumpkg",
            "__env__": "base",
            "__sls__": "base",
        },
    ]

    # Setup the State object
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        state_obj.load_modules(chunks[-1])
        state_obj.opts["state_aggregate"] = True  # Ensure state aggregation is enabled

        # Process each chunk with _mod_aggregate to simulate state execution
        state_obj.call_chunks(chunks)

        first_state_low = chunks[0]
        last_state_low = chunks[-1]
        # Verify that the requisites got aggregated as well
        assert first_state_low["require"] == last_state_low["require"]


def test_verify_onlyif_cmd_opts_exclude(minion_opts):
    """
    Verify cmd.run state arguments are properly excluded from cmd.retcode
    when passed.
    """
    low_data = {
        "onlyif": "somecommand",
        "cmd_opts_exclude": ["shell"],
        "cwd": "acwd",
        "root": "aroot",
        "env": [{"akey": "avalue"}],
        "prepend_path": "apath",
        "umask": "0700",
        "success_retcodes": 1,
        "timeout": 5,
        "runas": "doesntexist",
        "name": "echo something",
        "shell": "/bin/dash",
        "state": "cmd",
        "__id__": "this is just a test",
        "fun": "run",
        "__env__": "base",
        "__sls__": "sometest",
        "order": 10000,
    }

    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        mock = MagicMock()
        with patch.dict(state_obj.functions, {"cmd.retcode": mock}):
            #  The mock handles the exception, but the runas dict is being passed as it would actually be
            return_result = state_obj._run_check(low_data)
            mock.assert_called_once_with(
                "somecommand",
                ignore_retcode=True,
                python_shell=True,
                cwd="acwd",
                root="aroot",
                runas="doesntexist",
                env=[{"akey": "avalue"}],
                prepend_path="apath",
                umask="0700",
                timeout=5,
                success_retcodes=1,
            )


@pytest.mark.parametrize("verifier", (salt.state.State, salt.state.Compiler))
@pytest.mark.parametrize(
    "high,err_msg",
    (
        (
            {"/some/file": {"file.managed": ["source:salt://bla"]}},
            "Too many functions declared in state '/some/file' in SLS 'sls'. Please choose one of the following: managed, source:salt://bla",
        ),
        (
            {"/some/file": {"file": ["managed", "source:salt://bla"]}},
            "Too many functions declared in state '/some/file' in SLS 'sls'. Please choose one of the following: managed, source:salt://bla",
        ),
    ),
)
def test_verify_high_too_many_functions_declared_error_message(
    high, err_msg, minion_opts, verifier
):
    """
    Ensure the error message when a list item of a state call is
    accidentally passed as a string instead of a single-item dict
    is more meaningful. Example:

    /some/file:
      file.managed:
        - source:salt://bla

    /some/file:
      file:
        - managed
        - source:salt://bla

    Issue #38098.
    """
    high[next(iter(high))]["__sls__"] = "sls"
    with patch("salt.state.State._gather_pillar"):
        if verifier is salt.state.Compiler:
            state_obj = verifier(minion_opts, [])
        else:
            state_obj = verifier(minion_opts)
        res = state_obj.verify_high(high)
        assert isinstance(res, list)
        assert any(err_msg in x for x in res)


def test_load_modules_pkg(minion_opts):
    """
    Test load_modules when using this state:
    nginx:
      pkg.installed:
        - provider: pacmanpkg
    """
    data = {
        "state": "pkg",
        "name": "nginx",
        "__sls__": "test",
        "__env__": "base",
        "__id__": "nginx",
        "provider": "pacmanpkg",
        "order": 10000,
        "fun": "installed",
    }
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        state_obj.load_modules(data)
        for func in [
            "pkg.available_version",
            "pkg.file_list",
            "pkg.group_diff",
            "pkg.group_info",
        ]:
            assert func in state_obj.functions


def test_load_modules_list(minion_opts):
    """
    Test load_modules when using providers in state
    as a list, with this state:
    nginx:
      pkg.installed:
        - provider:
          - cmd: cmdmod
    """
    data = {
        "state": "pkg",
        "name": "nginx",
        "__sls__": "test",
        "__env__": "base",
        "__id__": "nginx",
        "provider": [salt.state.HashableOrderedDict([("cmd", "cmdmod")])],
        "order": 10000,
        "fun": "installed",
    }
    with patch("salt.state.State._gather_pillar"):
        state_obj = salt.state.State(minion_opts)
        state_obj.load_modules(data)
        for func in ["cmd.exec_code", "cmd.run", "cmd.script"]:
            assert func in state_obj.functions


def test_load_modules_dict(minion_opts):
    """
    Test load_modules when providers is a dict, which is
    not valid. Testing this state:
    nginx:
      pkg.installed:
        - provider: {cmd: test}
    """
    data = {
        "state": "pkg",
        "name": "nginx",
        "__sls__": "test",
        "__env__": "base",
        "__id__": "nginx",
        "provider": salt.state.HashableOrderedDict([("cmd", "test")]),
        "order": 10000,
        "fun": "installed",
    }
    mock_raw_mod = MagicMock()
    patch_raw_mod = patch("salt.loader.raw_mod", mock_raw_mod)
    with patch("salt.state.State._gather_pillar"):
        with patch_raw_mod:
            state_obj = salt.state.State(minion_opts)
            state_obj.load_modules(data)
            mock_raw_mod.assert_not_called()


def test_check_refresh_grains(minion_opts):
    """
    Test check_refresh when using this state:
    grains_refresh:
      module.run:
       - name: saltutil.refresh_grains
       - reload_grains: true
    Ensure that the grains are loaded when reload_grains
    is set.
    """
    data = {
        "state": "module",
        "name": "saltutil.refresh_grains",
        "__sls__": "test",
        "__env__": "base",
        "__id__": "grains_refresh",
        "reload_grains": True,
        "order": 10000,
        "fun": "run",
    }
    ret = {
        "name": "saltutil.refresh_grains",
        "changes": {"ret": True},
        "comment": "Module function saltutil.refresh_grains executed",
        "result": True,
        "__sls__": "test",
        "__run_num__": 0,
    }
    mock_refresh = MagicMock()
    patch_refresh = patch("salt.state.State.module_refresh", mock_refresh)
    with patch("salt.state.State._gather_pillar"):
        with patch_refresh:
            state_obj = salt.state.State(minion_opts)
            state_obj.check_refresh(data, ret)
            mock_refresh.assert_called_once()
            assert "cwd" in state_obj.opts["grains"]


def test_check_refresh_pillar(minion_opts, caplog):
    """
    Test check_refresh when using this state:
    pillar_refresh:
      module.run:
       - name: saltutil.refresh_pillar
       - reload_pillar: true
    Ensure the pillar is refreshed.
    """
    data = {
        "state": "module",
        "name": "saltutil.refresh_pillar",
        "__sls__": "test",
        "__env__": "base",
        "__id__": "pillar_refresh",
        "reload_pillar": True,
        "order": 10000,
        "fun": "run",
    }
    ret = {
        "name": "saltutil.refresh_pillar",
        "changes": {"ret": False},
        "comment": "Module function saltutil.refresh_pillar executed",
        "result": False,
        "__sls__": "test",
        "__run_num__": 0,
    }
    mock_refresh = MagicMock()
    patch_refresh = patch("salt.state.State.module_refresh", mock_refresh)
    mock_pillar = MagicMock()
    patch_pillar = patch("salt.state.State._gather_pillar", mock_pillar)
    with patch_pillar, patch_refresh:
        with caplog.at_level(logging.DEBUG):
            state_obj = salt.state.State(minion_opts)
            state_obj.check_refresh(data, ret)
            mock_refresh.assert_called_once()
            assert "Refreshing pillar..." in caplog.text


def test_module_refresh_runtimeerror(minion_opts, caplog):
    """
    test module_refresh when runtimerror occurs
    """
    mock_importlib = MagicMock()
    mock_importlib.side_effect = RuntimeError("Error")
    patch_importlib = patch("importlib.reload", mock_importlib)
    patch_pillar = patch("salt.state.State._gather_pillar", return_value="")
    with patch_importlib, patch_pillar:
        state_obj = salt.state.State(minion_opts)
        state_obj.module_refresh()
        assert (
            "Error encountered during module reload. Modules were not reloaded."
            in caplog.text
        )


def test_module_refresh_typeerror(minion_opts, caplog):
    """
    test module_refresh when typeerror occurs
    """
    mock_importlib = MagicMock()
    mock_importlib.side_effect = TypeError("Error")
    patch_importlib = patch("importlib.reload", mock_importlib)
    patch_pillar = patch("salt.state.State._gather_pillar", return_value="")
    with patch_importlib, patch_pillar:
        state_obj = salt.state.State(minion_opts)
        state_obj.module_refresh()
        assert (
            "Error encountered during module reload. Modules were not reloaded."
            in caplog.text
        )



"""
test for pillar file_tree.py
"""

import os
import pathlib

import pytest

import salt.pillar.file_tree as file_tree
import salt.utils.files
import salt.utils.stringutils
from tests.support.mock import MagicMock, patch


@pytest.fixture
def minion_id():
    return "test-host"


@pytest.fixture
def base_pillar_content():
    return {"files": {"hostfile": b"base", "groupfile": b"base"}}


@pytest.fixture
def dev_pillar_content():
    return {
        "files": {
            "hostfile": b"base",
            "groupfile": b"dev2",
            "hostfile1": b"dev1",
            "groupfile1": b"dev1",
            "hostfile2": b"dev2",
        }
    }


@pytest.fixture
def parent_pillar_content():
    return {"files": {"hostfile": b"base", "groupfile": b"base", "hostfile2": b"dev2"}}


@pytest.fixture
def pillar_path(tmp_path):
    return tmp_path / "file_tree"


@pytest.fixture
def configure_loader_modules(tmp_path, minion_id, pillar_path):
    cachedir = tmp_path / "cachedir"
    nodegroup_path = pathlib.Path("nodegroups", "test-group", "files")
    host_path = pathlib.Path("hosts", minion_id, "files")
    file_data = {
        (pillar_path / "base" / host_path / "hostfile"): "base",
        (pillar_path / "dev1" / host_path / "hostfile1"): "dev1",
        (pillar_path / "dev2" / host_path / "hostfile2"): "dev2",
        (pillar_path / "base" / nodegroup_path / "groupfile"): "base",
        (pillar_path / "dev1" / nodegroup_path / "groupfile1"): "dev1",
        (pillar_path / "dev2" / nodegroup_path / "groupfile"): "dev2",  # test merging
    }
    for filename in file_data:
        filename.parent.mkdir(parents=True, exist_ok=True)
        filename.write_text(salt.utils.stringutils.to_str(file_data[filename]))

    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value={"minions": [minion_id], "missing": []}),
    ):
        yield {
            file_tree: {
                "__opts__": {
                    "cachedir": cachedir,
                    "pillar_roots": {
                        "base": [str(pillar_path / "base")],
                        "dev": [
                            str(pillar_path / "base"),
                            str(pillar_path / "dev1"),
                            str(pillar_path / "dev2"),
                        ],
                        "parent": [
                            str(pillar_path / "base" / "sub1"),
                            str(pillar_path / "dev2" / "sub"),
                            str(pillar_path / "base" / "sub2"),
                        ],
                    },
                    "pillarenv": "base",
                    "nodegroups": {"test-group": [minion_id]},
                    "optimization_order": [0, 1, 2],
                    "file_buffer_size": 262144,
                    "file_roots": {"base": "", "dev": "", "parent": ""},
                    "extension_modules": "",
                    "renderer": "yaml_jinja",
                    "renderer_blacklist": [],
                    "renderer_whitelist": [],
                }
            }
        }


def test_absolute_path(base_pillar_content, minion_id, pillar_path):
    """
    check file tree is imported correctly with an absolute path
    """
    absolute_path = pillar_path / "base"
    mypillar = file_tree.ext_pillar(minion_id, None, str(absolute_path))
    assert base_pillar_content == mypillar

    with patch.dict(file_tree.__opts__, {"pillarenv": "dev"}):
        mypillar = file_tree.ext_pillar(minion_id, None, absolute_path)
        assert base_pillar_content == mypillar


def test_relative_path(base_pillar_content, dev_pillar_content, minion_id):
    """
    check file tree is imported correctly with a relative path
    """
    mypillar = file_tree.ext_pillar(minion_id, None, ".")
    assert base_pillar_content == mypillar

    with patch.dict(file_tree.__opts__, {"pillarenv": "dev"}):
        mypillar = file_tree.ext_pillar(minion_id, None, ".")
        assert dev_pillar_content == mypillar


def test_parent_path(parent_pillar_content, minion_id):
    """
    check if file tree is merged correctly with a .. path
    """
    with patch.dict(file_tree.__opts__, {"pillarenv": "parent"}):
        mypillar = file_tree.ext_pillar(minion_id, None, "..")
        assert parent_pillar_content == mypillar


def test_no_pillarenv(minion_id, caplog):
    """
    confirm that file_tree yells when pillarenv is missing for a relative path
    """
    with patch.dict(file_tree.__opts__, {"pillarenv": None}):
        mypillar = file_tree.ext_pillar(minion_id, None, ".")
        assert {} == mypillar

        for record in caplog.records:
            if record.levelname == "ERROR" and "pillarenv is not set" in record.message:
                break
        else:
            raise AssertionError("Did not find error message")


def test_file_tree_bytes(pillar_path, minion_id, base_pillar_content):
    """
    test file_tree pillar returns bytes
    """
    absolute_path = os.path.join(pillar_path, "base")
    mypillar = file_tree.ext_pillar(minion_id, None, absolute_path)
    assert base_pillar_content == mypillar

    with patch.dict(file_tree.__opts__, {"pillarenv": "dev"}):
        mypillar = file_tree.ext_pillar(minion_id, None, absolute_path)
        assert mypillar["files"]["groupfile"] == b"base"



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import copy
import datetime
import logging

import pytest

import salt.config
import salt.utils.schedule
from salt.utils.schedule import Schedule
from tests.support.mock import MagicMock, patch

# pylint: disable=import-error,unused-import
try:
    import croniter

    _CRON_SUPPORTED = True
except ImportError:
    _CRON_SUPPORTED = False
# pylint: enable=import-error

log = logging.getLogger(__name__)


# pylint: disable=too-many-public-methods,invalid-name
# delete_job tests
@pytest.mark.slow_test
def test_delete_job_exists(schedule):
    """
    Tests ensuring the job exists and deleting it
    """
    schedule.opts.update({"schedule": {"foo": "bar"}, "pillar": {}})
    assert "foo" in schedule.opts["schedule"]

    schedule.delete_job("foo")
    assert "foo" not in schedule.opts["schedule"]


@pytest.mark.slow_test
def test_delete_job_in_pillar(schedule):
    """
    Tests ignoring deletion job from pillar
    """
    schedule.opts.update({"pillar": {"schedule": {"foo": "bar"}}, "schedule": {}})
    assert "foo" in schedule.opts["pillar"]["schedule"]
    schedule.delete_job("foo")

    assert "foo" in schedule.opts["pillar"]["schedule"]


@pytest.mark.slow_test
def test_delete_job_intervals(schedule):
    """
    Tests removing job from intervals
    """
    schedule.opts.update({"pillar": {}, "schedule": {}})
    schedule.intervals = {"foo": "bar"}
    schedule.delete_job("foo")
    assert "foo" not in schedule.intervals


@pytest.mark.slow_test
def test_delete_job_prefix(schedule):
    """
    Tests ensuring jobs exists and deleting them by prefix
    """
    schedule.opts.update(
        {"schedule": {"foobar": "bar", "foobaz": "baz", "fooboo": "boo"}, "pillar": {}}
    )
    ret = copy.deepcopy(schedule.opts)
    del ret["schedule"]["foobar"]
    del ret["schedule"]["foobaz"]
    schedule.delete_job_prefix("fooba")
    assert schedule.opts == ret


@pytest.mark.slow_test
def test_delete_job_prefix_in_pillar(schedule):
    """
    Tests ignoring deletion jobs by prefix from pillar
    """
    schedule.opts.update(
        {
            "pillar": {"schedule": {"foobar": "bar", "foobaz": "baz", "fooboo": "boo"}},
            "schedule": {},
        }
    )
    ret = copy.deepcopy(schedule.opts)
    schedule.delete_job_prefix("fooba")
    assert schedule.opts == ret


# add_job tests
def test_add_job_data_not_dict(schedule):
    """
    Tests if data is a dictionary
    """
    data = "foo"
    pytest.raises(ValueError, Schedule.add_job, schedule, data)


def test_add_job_multiple_jobs(schedule):
    """
    Tests if more than one job is scheduled at a time
    """
    data = {"key1": "value1", "key2": "value2"}
    pytest.raises(ValueError, Schedule.add_job, schedule, data)


@pytest.mark.slow_test
def test_add_job(schedule):
    """
    Tests adding a job to the schedule
    """
    data = {"foo": {"bar": "baz"}}
    ret = copy.deepcopy(schedule.opts)
    ret.update(
        {
            "schedule": {
                "foo": {"bar": "baz", "enabled": True},
                "hello": {"world": "peace", "enabled": True},
            },
            "pillar": {},
        }
    )
    schedule.opts.update(
        {"schedule": {"hello": {"world": "peace", "enabled": True}}, "pillar": {}}
    )
    Schedule.add_job(schedule, data)
    assert schedule.opts == ret


# enable_job tests
@pytest.mark.slow_test
def test_enable_job(schedule):
    """
    Tests enabling a job
    """
    schedule.opts.update({"schedule": {"name": {"enabled": "foo"}}})
    Schedule.enable_job(schedule, "name")
    assert schedule.opts["schedule"]["name"]["enabled"]


@pytest.mark.slow_test
def test_enable_job_pillar(schedule):
    """
    Tests ignoring enable a job from pillar
    """
    schedule.opts.update({"pillar": {"schedule": {"name": {"enabled": False}}}})
    Schedule.enable_job(schedule, "name", persist=False)
    assert not schedule.opts["pillar"]["schedule"]["name"]["enabled"]


# disable_job tests
@pytest.mark.slow_test
def test_disable_job(schedule):
    """
    Tests disabling a job
    """
    schedule.opts.update({"schedule": {"name": {"enabled": "foo"}}, "pillar": {}})
    Schedule.disable_job(schedule, "name")
    assert not schedule.opts["schedule"]["name"]["enabled"]


@pytest.mark.slow_test
def test_disable_job_pillar(schedule):
    """
    Tests ignoring disable a job in pillar
    """
    schedule.opts.update(
        {"pillar": {"schedule": {"name": {"enabled": True}}}, "schedule": {}}
    )
    Schedule.disable_job(schedule, "name", persist=False)
    assert schedule.opts["pillar"]["schedule"]["name"]["enabled"]


# modify_job tests
@pytest.mark.slow_test
def test_modify_job(schedule):
    """
    Tests modifying a job in the scheduler
    """
    schedule_dict = {"foo": "bar"}
    schedule.opts.update({"schedule": {"name": "baz"}, "pillar": {}})
    ret = copy.deepcopy(schedule.opts)
    ret.update({"schedule": {"name": {"foo": "bar"}}})
    Schedule.modify_job(schedule, "name", schedule_dict)
    assert schedule.opts == ret


def test_modify_job_not_exists(schedule):
    """
    Tests modifying a job in the scheduler if jobs not exists
    """
    schedule_dict = {"foo": "bar"}
    schedule.opts.update({"schedule": {}, "pillar": {}})
    ret = copy.deepcopy(schedule.opts)
    ret.update({"schedule": {"name": {"foo": "bar"}}})
    Schedule.modify_job(schedule, "name", schedule_dict)
    assert schedule.opts == ret


def test_modify_job_pillar(schedule):
    """
    Tests ignoring modification of job from pillar
    """
    schedule_dict = {"foo": "bar"}
    schedule.opts.update({"schedule": {}, "pillar": {"schedule": {"name": "baz"}}})
    ret = copy.deepcopy(schedule.opts)
    Schedule.modify_job(schedule, "name", schedule_dict, persist=False)
    assert schedule.opts == ret


# enable_schedule tests
@pytest.mark.slow_test
def test_enable_schedule(schedule):
    """
    Tests enabling the scheduler
    """
    with patch(
        "salt.utils.schedule.Schedule.persist", MagicMock(return_value=None)
    ) as persist_mock:
        schedule.opts.update({"schedule": {"enabled": "foo"}, "pillar": {}})
        Schedule.enable_schedule(schedule)
        assert schedule.opts["schedule"]["enabled"]

    persist_mock.assert_called()


# disable_schedule tests
@pytest.mark.slow_test
def test_disable_schedule(schedule):
    """
    Tests disabling the scheduler
    """
    with patch(
        "salt.utils.schedule.Schedule.persist", MagicMock(return_value=None)
    ) as persist_mock:
        schedule.opts.update({"schedule": {"enabled": "foo"}, "pillar": {}})
        Schedule.disable_schedule(schedule)
        assert not schedule.opts["schedule"]["enabled"]

    persist_mock.assert_called()


# reload tests
def test_reload_update_schedule_key(schedule):
    """
    Tests reloading the schedule from saved schedule where both the
    saved schedule and schedule.opts contain a schedule key
    """
    saved = {"schedule": {"foo": "bar"}}
    ret = copy.deepcopy(schedule.opts)
    ret.update({"schedule": {"foo": "bar", "hello": "world"}})
    schedule.opts.update({"schedule": {"hello": "world"}})
    Schedule.reload(schedule, saved)
    assert schedule.opts == ret


def test_reload_update_schedule_no_key(schedule):
    """
    Tests reloading the schedule from saved schedule that does not
    contain a schedule key but schedule.opts does
    """
    saved = {"foo": "bar"}
    ret = copy.deepcopy(schedule.opts)
    ret.update({"schedule": {"foo": "bar", "hello": "world"}})
    schedule.opts.update({"schedule": {"hello": "world"}})
    Schedule.reload(schedule, saved)
    assert schedule.opts == ret


def test_reload_no_schedule_in_opts(schedule):
    """
    Tests reloading the schedule from saved schedule that does not
    contain a schedule key and neither does schedule.opts
    """
    saved = {"foo": "bar"}
    ret = copy.deepcopy(schedule.opts)
    ret["schedule"] = {"foo": "bar"}
    schedule.opts.pop("schedule", None)
    Schedule.reload(schedule, saved)
    assert schedule.opts == ret


def test_reload_schedule_in_saved_but_not_opts(schedule):
    """
    Tests reloading the schedule from saved schedule that contains
    a schedule key, but schedule.opts does not
    """
    saved = {"schedule": {"foo": "bar"}}
    ret = copy.deepcopy(schedule.opts)
    ret["schedule"] = {"foo": "bar"}
    schedule.opts.pop("schedule", None)
    Schedule.reload(schedule, saved)
    assert schedule.opts == ret


# eval tests
def test_eval_schedule_is_not_dict(schedule):
    """
    Tests eval if the schedule is not a dictionary
    """
    schedule.opts.update({"schedule": "", "pillar": {"schedule": {}}})
    pytest.raises(ValueError, Schedule.eval, schedule)


def test_eval_schedule_is_not_dict_in_pillar(schedule):
    """
    Tests eval if the schedule from pillar is not a dictionary
    """
    schedule.opts.update({"schedule": {}, "pillar": {"schedule": ""}})
    pytest.raises(ValueError, Schedule.eval, schedule)


def test_eval_schedule_time(schedule):
    """
    Tests eval if the schedule setting time is in the future
    """
    schedule.opts.update({"pillar": {"schedule": {}}})
    schedule.opts.update(
        {"schedule": {"testjob": {"function": "test.true", "seconds": 60}}}
    )
    now = datetime.datetime.now()
    schedule.eval()
    assert schedule.opts["schedule"]["testjob"]["_next_fire_time"] > now


def test_eval_schedule_time_eval(schedule):
    """
    Tests eval if the schedule setting time is in the future plus splay
    """
    schedule.opts.update({"pillar": {"schedule": {}}})
    schedule.opts.update(
        {"schedule": {"testjob": {"function": "test.true", "seconds": 60, "splay": 5}}}
    )
    now = datetime.datetime.now()
    schedule.eval()
    assert schedule.opts["schedule"]["testjob"]["_splay"] - now > datetime.timedelta(
        seconds=60
    )


@pytest.mark.skipif(not _CRON_SUPPORTED, reason="croniter module not installed")
def test_eval_schedule_cron(schedule):
    """
    Tests eval if the schedule is defined with cron expression
    """
    schedule.opts.update({"pillar": {"schedule": {}}})
    schedule.opts.update(
        {"schedule": {"testjob": {"function": "test.true", "cron": "* * * * *"}}}
    )
    now = datetime.datetime.now()
    schedule.eval()
    assert schedule.opts["schedule"]["testjob"]["_next_fire_time"] > now


@pytest.mark.skipif(not _CRON_SUPPORTED, reason="croniter module not installed")
def test_eval_schedule_cron_splay(schedule):
    """
    Tests eval if the schedule is defined with cron expression plus splay
    """
    schedule.opts.update({"pillar": {"schedule": {}}})
    schedule.opts.update(
        {
            "schedule": {
                "testjob": {"function": "test.true", "cron": "* * * * *", "splay": 5}
            }
        }
    )
    schedule.eval()
    assert (
        schedule.opts["schedule"]["testjob"]["_splay"]
        > schedule.opts["schedule"]["testjob"]["_next_fire_time"]
    )


@pytest.mark.slow_test
def test_handle_func_schedule_minion_blackout(schedule):
    """
    Tests eval if the schedule from pillar is not a dictionary
    """
    schedule.opts.update({"pillar": {"schedule": {}}})
    schedule.opts.update({"grains": {"minion_blackout": True}})

    schedule.opts.update(
        {"schedule": {"testjob": {"function": "test.true", "seconds": 60}}}
    )
    data = {
        "function": "test.true",
        "_next_scheduled_fire_time": datetime.datetime(2018, 11, 21, 14, 9, 53, 903438),
        "run": True,
        "name": "testjob",
        "seconds": 60,
        "_splay": None,
        "_seconds": 60,
        "jid_include": True,
        "maxrunning": 1,
        "_next_fire_time": datetime.datetime(2018, 11, 21, 14, 8, 53, 903438),
    }

    with patch.object(salt.utils.schedule, "log") as log_mock:
        with patch("salt.utils.process.daemonize"), patch("sys.platform", "linux2"):
            schedule.handle_func(False, "test.ping", data)
            assert log_mock.exception.called


def test_handle_func_check_data(schedule):
    """
    Tests handle_func to ensure that __pub_fun_args is not
    being duplicated in the value of kwargs in data.
    """

    data = {
        "function": "test.arg",
        "_next_scheduled_fire_time": datetime.datetime(2018, 11, 21, 14, 9, 53, 903438),
        "run": True,
        "args": ["arg1", "arg2"],
        "kwargs": {"key1": "value1", "key2": "value2"},
        "name": "testjob",
        "seconds": 60,
        "_splay": None,
        "_seconds": 60,
        "jid_include": True,
        "maxrunning": 1,
        "_next_fire_time": datetime.datetime(2018, 11, 21, 14, 8, 53, 903438),
    }

    with patch("salt.utils.process.daemonize"), patch("sys.platform", "linux2"):
        with patch.object(schedule, "standalone", return_value=True):
            # run handle_func once
            schedule.handle_func(False, "test.arg", data)

            # run handle_func and ensure __pub_fun_args
            # is not in kwargs
            schedule.handle_func(False, "test.arg", data)

            assert "kwargs" in data
            assert "__pub_fun_args" not in data["kwargs"]


def test_handle_func_check_dicts(schedule):
    """
    Tests that utils, functions, and returners dicts are not
    empty after handle_func has run on Windows.
    """

    data = {
        "function": "test.arg",
        "_next_scheduled_fire_time": datetime.datetime(2018, 11, 21, 14, 9, 53, 903438),
        "run": True,
        "args": ["arg1", "arg2"],
        "kwargs": {"key1": "value1", "key2": "value2"},
        "name": "testjob",
        "seconds": 60,
        "_splay": None,
        "_seconds": 60,
        "jid_include": True,
        "maxrunning": 1,
        "_next_fire_time": datetime.datetime(2018, 11, 21, 14, 8, 53, 903438),
    }

    with patch("salt.utils.process.daemonize"):
        with patch.object(schedule, "standalone", return_value=True):
            # simulate what happens before handle_func is called on Windows
            schedule.functions = {}
            schedule.returners = {}
            schedule.utils = {}
            schedule.handle_func(False, "test.arg", data)

            assert schedule.functions != {}
            assert schedule.returners != {}
            assert schedule.utils != {}
